import numpy as np
import matplotlib.pyplot as plt

from ..utils.exceptions import user_warning, InputError


class Bishop:
    """ Bishop slip circles

    Class for the computation of the factor of safety against slip
    failure. This factor of safety is computed with the following
    equation (Verruijt, 2012):

    .. math::
       F=\\frac{\\sum \\frac{c+(\\gamma h-p) \\tan \\phi}{\\cos \\alpha(
       1+\\tan \\alpha \\tan \\phi / F)}}{\\sum \\gamma h \\sin \\alpha}

    The top of the soil is defined by one point, point2, see the Figure.
    All slip circles must go through (0,0) and point2, this means that
    if a custom slip circle is defined this circle must go through both
    points. If a custom slip circle is not given, slip circles are
    automatically generated. These circles are generated with an interval
    of :py:obj:`y_step`.

    .. figure:: _figures/bishop-def.png
       :align: center
       :alt: definition of the input parameters for the Bishop class

    Parameters
    ----------
    point2 : tuple
        x and y coordinate of point2
    y_step : float, optional, default: 1
        step size of the y coordinate for generating circles, used if a
        custom circle is not defined
    wlev : float, optional, defautl: None
        y coordinate of the water level, by default the water level is
        None, which means that there is no water
    SlipCircle : :py:class:`SlipCircle`
        user defined custom slip circle, circle must be defined with
        :py:class:`SlipCircle`

    Attributes
    ----------
    x2, y2 : float
        x and y coordinate of point2
    wlev : float
        y coordinate of the water level
    circles : dict
        dictionary with the centre and radius of the circle. Slices are
        added to the dict when generated during the computation.
    layers : dict
        dictionary with the defined layers
    normative : int
        id of the normative :py:class:`SlipCircle`
    """

    def __init__(self, point2, y_step=1, wlev=None, SlipCircle=None):
        """ See help(Bishop) for more info """
        # set point2 and y_step as attributes
        self.x2 = point2[0]
        self.y2 = point2[1]
        self.wlev = wlev

        # check if a custom circle has been specified
        if SlipCircle is None:
            # no circle given, generate all circles
            self.circles = self._make_circles(y_step)
        else:
            # set custom circle as attribute with id 0
            self.circles = {1: SlipCircle}

        # make attribute for the layers
        self.layers = {}

        # set attribute to store id of normative slip circle
        self.normative = None

    def _make_circles(self, y_step):
        """ Compute all possible circles

        This function computes all possible circles going through two
        points, the first is (0,0) and the second must be specified as
        an argument. The possible circles are computed with the following
        constraints:

        .. math::
           y \\geq y_{2}
           0 \\leq x \\leq x_{2}

        Parameters
        ----------
        y_step : float, optional, default: 1
            step size with which the y coordinate is increased

        Returns
        -------
        dict
            dictionary with the centre coordinate, radius and angle of the
            arc for all possible circles
        """
        # make dict to store generated circles
        circles = {}

        # set start value for y of the centre of the circle
        y = self.y2

        # set variable for the circle id
        id = 1

        # generate circles in the domain
        while True:
            # compute x coordinate of the centre of the circle
            x = (self.x2**2 + self.y2**2 - 2*self.y2*y)/(2*self.x2)

            # compute radius of the circle
            r = np.sqrt(x**2+y**2)

            # compute angle of the arc
            chord = np.sqrt(self.x2**2 + self.y2**2)

            # make circle object and add to dict
            circles[id] = SlipCircle(centre=(x, y), r=r)

            # check if x is negative, break criteria
            if x <= 0:
                # break loop
                break
            else:
                # increase y with the step size and increment i
                y += y_step
                id += 1

        return circles

    def add_layer(
            self, gamma, c, phi, name, ymin=None, ymax=None, gamma_sat=None):
        """ Add layer to the soil

        Add a soil layer for the computation. In case of a homogeneous
        soil the arguments :py:obj:`ymin` and :py:obj:`ymax` do not have
        to be specified. However, if the soil consists of several layers
        these parameters must be given for each layer.

        .. warning::
           soil layers must be added sequentially from the lowest layer
           to the highest layer.

        Parameters
        ----------
        gamma : float
            volumetric weight of the material [kN/m³]
        c : float
            cohesion of the material [kPa]
        phi : float
            internal friction angle [deg]
        name : str
            name of the layer
        ymin : float, optional, default: None
            y coordinate of the start of the layer, required if more
            than 1 layer is added
        ymax : float, optional, default: None
            y coordinate of the end of the layer, required if more
            than 1 layer is added
        gamma_sat : float, optional, default: None
            saturated volumetric weight of the material [kN/m³]. Only
            required if a :py:attr:`wlev` is specified.

        Raises
        ------
        InputError
            if more than 1 layer is specified and ymin and/or ymax has
            not been specified or gamma_sat is not given when a wlev has
            been specified
        """
        # convert phi to radians
        phi = phi*np.pi/180

        # check if name is already in
        if name in self.layers.keys():
            # show warning that layer is overwritten
            user_warning(
                (f'{name} already a layer, layer has been overwritten by the'
                  ' specified layer'))

        # check if a wlev has been defined, and a gamma_sat
        if self.wlev is not None and gamma_sat is None:
            raise InputError(
                'gamma_sat must be given when a wlev is specified')

        # add to layers attribute
        self.layers[name] = {
            'gamma': gamma, 'gamma_sat': gamma_sat, 'c': c, 'phi': phi,
            'y_coord': (ymin, ymax)}

        # check if layers are correctly added
        if self.layers:
            # a layer has already been added, iterate over the layers
            # since layers must be added sequentially
            for i, (layer, params) in enumerate(self.layers.items()):
                # get y coordinates of the layer
                ymin_current, ymax_current = params['y_coord']

                # start checking for more than 1 layer
                if i > 0:
                    # next layer, check if ymax of the previous layer
                    # is equal to ymin of the this layer and ymax is
                    # larger since layer must be on top of the previous
                    # layer
                    if (ymax_previous == ymin_current
                            and ymax_current > ymax_previous):
                        # correct
                        pass
                    else:
                        raise InputError(
                            (f'layer {layer} has invalid coordinates, layers '
                              'must be added sequentially'))

                # save coordinates of the current layer
                ymin_previous, ymax_previous = ymin_current, ymax_current

    def compute(self, num_slices, max_iter=50, ftol=0.05, gamma_w=None):
        """ Compute the factor of safety

        Method to compute the factor of safety for all generated
        circles, or the specified circle. Updates :py:attr:`normative`
        with the id of the normative slip circle, i.e. the slip circle
        with the lowest factor of safety.

        Parameters
        ----------
        num_slices : int
            number of slices
        max_iter : int, optional, default: 50
            maximum number of iterations
        ftol : float, optional, default: 0.05
            break criterium, when the change between the previous and
            the current factor of safety is below the change the
            iteration is ended
        gamma_w : float, optional, default: None
            volumetric weight of water [kN/m³]
        """
        # check if max_iter is an int
        if not isinstance(max_iter, int):
            raise InputError('max_iter must be an int')

        # set variable to track normative F
        F_norm = 10*10

        # iterate over the circles
        for id, SlipCircle in self.circles.items():
            # check if slices have alreay been made
            if SlipCircle.slices is None:
                # slices not have not yet been made, so make slices
                SlipCircle.make_slices(
                    num_slices=num_slices, point2=(self.x2, self.y2),
                    layers=self.layers)

            # set first estimate for F and iteration tracker
            F = 1
            i = 0

            # iteratively compute the factor of safety
            while True:
                # set strength and load variable
                strength, load = 0, 0

                # iterate over the slices
                for slice, coords in SlipCircle.slices.items():
                    # check slice number
                    if slice == 0:
                        # only include pressure for the first slice
                        pressure = True
                    else:
                        # other slides no pressure, otherwise doubling
                        pressure = False

                    # compute load and strength
                    load += self._load(coords['alpha_s'], coords['h'])
                    strength += self._strength(
                        alpha_s=coords['alpha_s'], heights=coords['h'], F=F,
                        pressure=pressure, gamma_w=gamma_w)

                # # compute factor of safety
                F_new = strength/load

                # check for break criteria
                if np.abs(F_new - F) <= ftol or i == (max_iter-1):
                    # set F as attribute of the slip circle
                    SlipCircle.F = F_new

                    # check if computed F is smaller than current normative
                    if F_new <= F_norm:
                        # update F_norm and set id as normative
                        F_norm = F_new
                        self.normative = id

                    # break computation
                    break

                # update F and increment i
                F = F_new
                i += 1

    def _load(self, alpha_s, heights):
        """ Compute the load on the slice

        Method computes the loading force of a slice with the following
        equation:

        .. math::
           \\text{load} = \\gamma h \\sin \\alpha

        Parameters
        ----------
        alpha_s : float
            slip angle of the slice [rad]
        heights : dict
            dictionary with the height of each layer

        Returns
        -------
        float
            load of the slice
        """
        # set variable for adding weights
        W = 0

        # iterate over the layers
        for layer, h in heights.items():
            # check if there is a wlev
            if self.wlev is not None:
                # check if layer is above or below the wlev
                if self.wlev >= np.max(h):
                    # entire layer is below the water level
                    W += self.layers[layer]['gamma_sat']*(np.max(h)-np.min(h))

                elif np.min(h) >= self.wlev:
                    # entire layer is above the wlev
                    W += self.layers[layer]['gamma']*(np.max(h)-np.min(h))

                else:
                    # layer is partly in the water, compute wet and dry h
                    h_dry = np.max(h) - self.wlev
                    h_wet = (np.max(h)-np.min(h)) - h_dry

                    # compute weight of the layer
                    W += (self.layers[layer]['gamma_sat']*h_wet
                          + self.layers[layer]['gamma']*h_dry)

            else:
                # not wlev specified, compute the load
                W += self.layers[layer]['gamma']*(np.max(h)-np.min(h))

        return W*np.sin(alpha_s)

    def _strength(self, alpha_s, heights, F, pressure, gamma_w=None):
        """ Compute the strength of a slice

        Method computes the loading force of a slice with the following
        equation:

        .. math::
           \\text{strength} = \\sum \\frac{c+(\\gamma h-p) \\tan \\phi}
           {\\cos \\alpha(1+\\tan \\alpha \\tan \\phi / F)}

        Parameters
        ----------
        alpha_s : float
            slip angle of the slice [rad]
        heights : dict
            dictionary with the height of each layer
        F : float
            factor of safety [-]
        pressure : bool
            True if the pressure must be included in the computation,
            False if not
        gamma_w : float, optional, default: None
            volumetric weight of water [kN/m³]

        Returns
        -------
        float
            strength of the slice
        """
        # set variable for adding weights
        W = 0

        # check if a wlev is defined
        if self.wlev is not None and pressure:
            # check if gamma_w has been given
            if gamma_w is None:
                # no volumetric weight of water, raise error
                raise InputError(
                    'gamma_w must be specified when computing with a wlev')

            # get top and bottom coordinate of the slice
            ytop = np.max(list(heights.values()))
            ybottom = np.min(list(heights.values()))

            # check if top is above the wlev
            if ytop >= self.wlev:
                # compute pressure in the middle of the slice
                p = 0.5*(self.wlev-ybottom)*gamma_w
            else:
                # top of the slice is not above the wlev
                # compute pressure in the middle of the slice
                p = gamma_w*(self.wlev-ytop+0.5*(ytop-ybottom))

        else:
            # no wlev, thus pressure is zero
            p = 0

        # iterate over the layers
        for i, (layer, h) in enumerate(heights.items()):
            # check if first layer
            if i == 0:
                # first layer of the slice offers the resistance
                # against sliding, thus save parameters of this layer
                params = self.layers[layer]

            # check if there is a wlev
            if self.wlev is not None:
                # check if layer is above or below the wlev
                if self.wlev >= np.max(h):
                    # entire layer is below the water level
                    W += self.layers[layer]['gamma_sat']*(np.max(h)-np.min(h))

                elif np.min(h) >= self.wlev:
                    # entire layer is above the wlev
                    W += self.layers[layer]['gamma']*(np.max(h)-np.min(h))

                else:
                    # layer is partly in the water, compute wet and dry h
                    h_dry = np.max(h) - self.wlev
                    h_wet = (np.max(h)-np.min(h)) - h_dry

                    # compute weight of the layer
                    W += (self.layers[layer]['gamma_sat']*h_wet
                          + self.layers[layer]['gamma']*h_dry)

            else:
                # not wlev specified, compute the weight of the layer
                W += params['gamma']*(np.max(h)-np.min(h))

        # compute the strength
        A = params['c'] + (W - p)*np.tan(params['phi'])
        B = np.tan(alpha_s)*np.tan(params['phi'])/F
        C = np.cos(alpha_s)*(1 + B)
        strength = A/C

        return strength

    def plot(self, id=None, show_slices=False):
        """ Plot slip circle(s)

        Method to plot the slip circle(s)

        Parameters
        ----------
        id : int, optional, default: None
            id of the slip circle to plot, by default all slip circles
            are plotted
        show_slices : bool, optional, default: False
            if the slices must be shown, only available after the
            computation as the slices are not generated before the
            computation. The default value is False, meaning that the
            slices will not be plotted.
        """
        # create figure
        fig, ax = plt.subplots()

        # define coordinates of the boundary
        x = [-self.x2, 0, self.x2, 2*self.x2]
        y = [0, 0, self.y2, self.y2]

        # plot the boundary
        ax.plot(x,y, zorder=20, color='k', lw=2)

        # check if custom id has been specified
        if id is not None:
            # plot one circle
            ax = self.circles[id]._make_figure(ax, show_slices)

            # compute ymax and ymin
            ymax = self.circles[id].xy[1] + self.circles[id].r
            ymin = self.circles[id].xy[1] - self.circles[id].r

            # check if ymin is zero
            if ymin >= -1 or ymin <= 1:
                ymin = -5
                
            # set title
            title = f'Slip circle id={id}'

        else:
            # set variables for ymax and ymin
            ymax, ymin = 0, 0

            # iterate over the generated circles to plot all circles
            for id, SlipCircle in self.circles.items():
                # plot circle
                ax = SlipCircle._make_figure(ax, show_slices=False)

                # compute ymax and ymin
                ymax1 = self.circles[id].xy[1] + self.circles[id].r
                ymin1 = self.circles[id].xy[1] - self.circles[id].r

                # check if largest value is larger than current largest
                if ymax1 > ymax:
                    ymax = ymax1

                # check if smallest value is smaller than current smallest
                if ymin1 < ymin:
                    ymin = ymin1

            # set title
            title = 'Slip circles'

        # check if more than one layer has been defined
        if len(self.layers) > 1:
            # plot the layers
            colors = ['g', 'r', 'c', 'm', 'y']
            for i, (layer, params) in enumerate(self.layers.items()):
                # check if out of range for colors
                if i == len(colors):
                    # reset i
                    i = 0

                # plot layer
                ax.axhline(
                    y=params['y_coord'][1], label=f'top of {layer}', ls='--',
                    color=colors[i])

            plt.legend()

        # check if a wlev has been specified
        if self.wlev is not None:
            # add wlev
            plt.axhline(y=self.wlev, color='blue')

        # set layout
        # ax.set_xlim(np.min(x), np.max(x))
        ax.set_ylim(ymin*1.1, ymax*1.1)
        ax.set_xlabel('x coordinate [m]')
        ax.set_ylabel('y coordinate [m]')
        ax.set_title(title)
        ax.grid()
        fig.gca().set_aspect('equal', adjustable='box')
        fig.tight_layout()
        plt.show()


class SlipCircle:
    """ Define a slip circle

    Define a custom slip circle to use in :py:class:`Bishop`

    Parameters
    ----------
    centre : tuple
        x and y coordinate of the centre of the circle
    r : float
        radius of the circle

    Attributes
    ----------
    xy : tuple
        x and y coordinate of the centre of the circle
    r : float
        radius of the circle
    F : float
        computed factor of safety
    slices : dict
        dictionary with the coordinates, height and slip angle of the
        slices
    """

    def __init__(self, centre, r):
        """ See help(SlipCircle) for more info """
        # set input as attributes
        self.xy = centre
        self.r = r

        # set attribute for factor of safety and slices
        self.F = None
        self.slices = None

    def __repr__(self):
        return f'SlipCircle(centre={self.xy}, r={np.round(self.r, 2)})'

    def __str__(self):
        return (f'Slip cricle with centre {self.xy} and a radius of '
                f'{np.round(self.r, 2)}')

    @staticmethod
    def _compute_heights(y1, y2, y3, y4, layers):
        """ Compute the height of the slice

        Method to compute the height of a slice. Note that the following
        condition must hold in the input:

        .. math::
        y_4 > y_3 > y_2 > y_1

        Parameters
        ----------
        y1, y2 : float
        lower y coordinates of the slice, where y2 must be larger
        than y1
        y3, y4 : float
        higher y coordinates of the slice, where y4 must be larger
        than y3

        """
        # create dict to store the heights
        height = {}

        # compute y coordinates of the middle of the circle
        ym_low = y1 + 0.5*(y2-y1)
        ym_top = y3 + 0.5*(y4-y3)

        # check if more than one layers has been defined
        if len(layers) > 1:
            # set variable to track progress through the layers
            # starting value of this variable is ym_low
            y = ym_low

            # iterate over the layers
            for layer, params in layers.items():
                # get ymin and ymax of the layer
                ymin, ymax = params['y_coord']

                # check if part of the slice in the current layer
                if ymax > y:
                    # check if top of the layer is higher than slice
                    if ymax > ym_top:
                        height[layer] = (ym_top, y)
                    else:
                        height[layer] = (ymax, y)

                        # set tracker to top of the current layer
                        y = ymax

        elif len(layers) == 1:
            # only one layer defined
            # get name of the layer
            name = list(layers.keys())[0]

            # add to dict
            height[name] = (ym_top, ym_low)

        else:
            # no layers have been defined, raise InputError
            raise InputError('No layers have been defined')

        return height

    def _get_intersect(self, x):
        """ Get the intersection between a slice and the circle

        method returns the y coordinate of the intersection point
        between the circle and the slice. Note that two y coordinates
        are possible, this method returns the one below the centre of
        the circle. This is, because the slip circle is always located
        below the centre of the circle

        Parameters
        ----------
        x : float
            x coordinate of the intersection

        Returns
        -------
        float
            y coordinate of the intersection, note that this is the
            lowest intersection point
        """
        # compute a, b and c for the quadratic formula
        a = 1
        b = -2*self.xy[1]
        c = self.xy[1]**2 - self.r**2 + (x-self.xy[0])**2

        # compute factor of the sqrt
        D = b**2 - 4*a*c

        # check value of D
        if D > 0:
            # two solutions possible, compute y coordinates
            y1 = (-b + np.sqrt(b**2 - 4*a*c))/(2*a)
            y2 = (-b - np.sqrt(b**2 - 4*a*c))/(2*a)

            # return lowest value
            return np.min([y1, y2])

        elif D == 0 or D > -0.1**10:
            # one solutions
            return -b/(2*a)

        else:
            # negative value, imaginary solutions, raise error
            raise ValueError(
                'When making slices an imaginary solutions is encountered')

    def _angle(self, yc_inter, x1, y1, x2, y2):
        """ Compute the slip angle of a slice

        Method computes the slip angle, :math:`\\alpha_s`, of a slice
        with respect to the centre of the slice

        Parameters
        ----------
        yc_inter : float
            y coordinate of the intersection between the circle and a
            line vertical from the centre of the circle
        x1, y1 : float
            x and y coordinate of the lower left point of the slice
        x2, y2 : float
            x and y coordinate of the lower right point of the slice

        Returns
        -------
        float
            slip angle of the slice [rad]
        """
        # compute coordinate of the middle of the slice
        xm = x1 + 0.5*(x2-x1)
        ym = y1 + 0.5*(y2-y1)

        # check if xm is equal to xc
        if xm == self.xy[0]:
            # angle is zero
            return 0
        else:
            # compute chord from the line vertical from the middle of
            # the slice to the middle of the slice
            chord = np.sqrt((xm-self.xy[0])**2 + (ym-yc_inter)**2)

            # compute slip angle of the middle of the slice
            alpha_s = 2*np.arcsin(0.5*chord/self.r)

            # check if right or left from the centre of the circle
            if xm < self.xy[0]:
                # left of the centre, angle must be negative
                return -alpha_s
            elif xm > self.xy[0]:
                # right from the centre, positive angle
                return alpha_s

    def _make_figure(self, ax, show_slices):
        """ Method to generate a figure of the circle """
        # plot circle
        circle1 = plt.Circle(
            self.xy, self.r, color='darkgrey', ls='--', fill=False)
        ax.add_artist(circle1)

        # plot centre of the circle
        ax.plot(self.xy[0], self.xy[1], marker='.', color='darkgrey')

        # check if slices must be shown
        if self.slices is not None and show_slices:
            # iterate over the slices
            for slice, coord in self.slices.items():
                # plot the slice
                ax.plot(coord['x'], coord['y'], color='k')

        elif self.slices is not None and show_slices:
            # slices have not been generated but are requested
            user_warning(
                ('show_slices is True, but the slices have not yet been '
                 'generated'))

        return ax

    def make_slices(self, num_slices, point2, layers):
        """ Make slices

        Method to divide the given circle into slices

        Parameters
        ----------
        num_slices : int
            number of slices
        circle : dict
            dictionary with the centre and radius of the circle
        layers : dict
            dictionary with the layers
        """
        # check if number of slices is an integer
        if not isinstance(num_slices, int):
            raise InputError('Slices must be an int')

        # compute width of one slice
        width = point2[0]/num_slices

        # set dict to store the coordinates of the slices
        slices = {}

        # add first slice
        # compute intersection with circle (rightmost intersection)
        y_inter = self._get_intersect(x=width)

        # compute y coordinate of the intersection between a line from
        # the centre of the circle and the circle
        yc_inter = self._get_intersect(x=self.xy[0])

        # compute slip angle of the first slice
        alpha_s = self._angle(yc_inter, 0, 0, width, y_inter)

        # compute the height of the slice
        heights = self._compute_heights(
            0, y_inter, point2[1]*width/point2[0], 0, layers)

        # add to dict
        slices[0] = {
            'x': [0, width, width, 0],
            'y': [0, y_inter, point2[1]*width/point2[0], 0],
            'alpha_s': alpha_s,
            'h': heights}

        # set rightmost intersection as leftmost intersection for
        # the next slice
        y_inter_1 = y_inter

        # compute the coordinates of each slice
        for slice in range(1, num_slices):
            # compute coordinates of the slice
            x1 = width*slice
            x2 = x1 + width
            y1 = point2[1]*x1/point2[0]
            y2 = point2[1]*x2/point2[0]

            # get intersection with the circle
            y_inter_2 = self._get_intersect(x=x2)

            # compute slip angle of the slice
            alpha_s = self._angle(yc_inter, x1, y_inter_1, x2, y_inter_2)

            # compute the height of the slice
            heights = self._compute_heights(
                y_inter_1, y_inter_2, y2, y1, layers)

            # add coordinates and slip angle to dict
            slices[slice] = {
                'x': [x1, x2, x2, x1, x1],
                'y': [y_inter_1, y_inter_2, y2, y1, y_inter_1],
                'alpha_s': alpha_s,
                'h': heights}

            # set rightmost intersection as leftmost intersection for
            # the next slice
            y_inter_1 = y_inter_2

        # set slices as attribute
        self.slices = slices

    def plot(self, show_slices=False):
        """ Plot the circle

        Parameters
        ----------
        show_slices : bool, optional, default: False
            if the slices must be shown, only available after the
            computation as the slices are not generated before the
            computation. The default value is False, meaning that the
            slices will not be plotted.
        """
        # set additional spacing
        spacing = self.r/5
        # create figure
        fig, ax = plt.subplots()

        # add circle
        ax = self._make_figure(ax, show_slices)

        # set layout
        ax.set_ylim(self.xy[1]-self.r-spacing, self.xy[1]+self.r+spacing)
        ax.set_xlim(self.xy[0]-self.r-spacing, self.xy[0]+self.r+spacing)
        ax.set_xlabel('x coordinate [m]')
        ax.set_ylabel('y coordinate [m]')
        ax.set_title('Slip cricle')
        ax.grid()
        fig.gca().set_aspect('equal', adjustable='box')
        fig.tight_layout()
        plt.show()



import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import fsolve

from .utils.exceptions import InputError, limitstate_warning
from .utils.wave import dispersion
from .core.battjes import BattjesGroenendijk


class LimitState:
    """ Generate LimitState for design

    Define a design limit state, for instance the Ultimate Limit State
    (ULS) or Serviceability Limit State (SLS). The defined limit state[s]
    can then be given as arguments to design one of the supported
    breakwater types, see Table 1 for the required parameters for these
    design classes. A full list of possible arguments can be seen under
    Keyword Arguments.

    Table 1: Required parameters for each design class

    +---------------+------------+------------+------------+------------+
    | Parameter     |    RRM     |  CRM       |  RC        |       CC   |
    +===============+============+============+============+============+
    | h             | x          | x          | x          | x          |
    +---------------+------------+------------+------------+------------+
    | q             | x          | x          | x          | x          |
    +---------------+------------+------------+------------+------------+
    | H13           | o :sup:`1` | o :sup:`1` | o :sup:`1` | o :sup:`1` |
    +---------------+------------+------------+------------+------------+
    | Hm0           | o :sup:`1` | o :sup:`1` | o :sup:`1` | o :sup:`1` |
    +---------------+------------+------------+------------+------------+
    | H2_per        | o :sup:`2` |            |            |            |
    +---------------+------------+------------+------------+------------+
    | Hmax          |            |            | x          | x          |
    +---------------+------------+------------+------------+------------+
    | Tm            | o :sup:`3` |            |            |            |
    +---------------+------------+------------+------------+------------+
    | T13           |            |            | o :sup:`3` | o :sup:`3` |
    +---------------+------------+------------+------------+------------+
    | T_m_min_1     | o :sup:`3` | o :sup:`3` | o :sup:`3` | o :sup:`3` |
    +---------------+------------+------------+------------+------------+
    | Nod           | o :sup:`4` | x          |            |            |
    +---------------+------------+------------+------------+------------+
    | Sd            | o :sup:`4` |            |            |            |
    +---------------+------------+------------+------------+------------+

    | :sup:`1` uses :py:meth:`get_Hs` to get Hs, so at least one of the
      wave heights must be given
    | :sup:`2` if H2_per not specified in the limit state it is computed
      with :py:meth:`get_H2`
    | :sup:`3` wave periods can be transformed in deep water with
      :py:meth:`transform_periods`
    | :sup:`4` either Nod or Sd is required, the other can be computed
      with :py:meth:`Nod` or :py:meth:`Sd`

    .. note::
       The design classes
       :obj:`bw.RockRubbleMound <breakwater.rubble.RockRubbleMound>`,
       :obj:`bw.ConcreteRubbleMound <breakwater.rubble.ConcreteRubbleMound>`
       and :obj:`bw.Caisson <breakwater.caisson.Caisson>` perform
       :sup:`1` and :sup:`2` automatically, but not :sup:`3` and :sup:`4`

    Parameters
    ----------
    h : float
        the water depth [m]
    label : str
        label to identify the LimitState, for instance ULS or SLS

    Keyword Arguments
    -----------------
    q : float
        mean overtopping discharge per meter structure width [l/s per m]
    Hs : float
        the significant wave height [m]
    H13 : float
        mean of the highest 1/3 of the wave heights [m]
    Hm0 : float
        the spectral wave height [m]
    Ho : float
        deep water wave height [m]
    H2_per : float
        the wave height exceeded by 2% of the waves [m]
    Hmax : float
        design wave height for the Goda formula, equal to the mean of
        the highest 1/250 of the wave heights [m].
    H1250 : float
        mean of the highest 1/250 of the wave heights [m], automatically
        interpreted as Hmax.
    Tm : float
        the mean wave period [s]
    T13 : float
         the significant wave period [s]
    Ts : float
         the significant wave period [s]
    T_m_min_1 : float
        :math:`T_{m-1.0}`, the energy wave period [s]
    Tp : float
        the peak period [s]
    Nod : float
        Damage number, used in the formula for the toe stability. Can
        also be computed with :py:meth:`Nod` [-]
    Sd : float
        Damage level parameter, used in Van der Meer formula. Can also
        be computed with :py:meth:`Sd` [-]


    Attributes
    ----------
    conditions : dict
        Dictionary with all parameters
    deep_water : bool
        True if deep water assumption is valid, False if not valid. The
        bool is computed with the Tm, Tp, T_m_min_1 or T13 (in this
        order)
    h : float
        the water depth [m]
    label : str
        identifier of the LimitState
    """
    def __init__(self, h, label, **kwargs):
        """ See help(LimitState) for more info """
        self.conditions = kwargs
        self.h = h
        self.label = label

        if 'H1250' in self.conditions:
            # H1/250 is equal to Hmax thus set Hmax
            self.conditions['Hmax'] = self.conditions['H1250']

        if 'Tm' in self.conditions:
            period = 'Tm'
        elif 'Tp' in self.conditions:
            period = 'Tp'
        elif 'T_m_min_1' in self.conditions:
            period = 'T_m_min_1'
        elif 'Ts' in self.conditions:
            self.conditions['T13'] = self.conditions['Ts']
            period = 'T13'
        elif 'T13' in self.conditions:
            period = 'T13'
        else:
            raise InputError(
                ('No wave period in LimitState, please specify (at least) one'
                 ' of the following wave periods: \'Tm\', \'Tp\', \'Ts\' '
                 '\'T_m_min_1\' or \'T13\'' ))

        if self.h/self.L(period=period) > 0.5:
            self.deep_water = True
        else:
            self.deep_water = False

    def __str__(self):
        return str(self.conditions)

    def __setitem__(self, key, value):
        self.conditions[key] = value

    def __getitem__(self, key):
        return self.conditions[key]

    def check_deep_water(self):
        """ Check if the deep water assumption is valid

        Deep water assumption is valid if h/L > 0.5, where L is computed
        with the dispersion relation using the mean period. Method
        updates the attribute :py:attr:`deep_water` based on the check.

        Raises
        ------
        KeyError
            If Tm is not in the LimitState
        """
        if self.h/self.L(period='Tm') > 0.5:
            self.deep_water = True
        else:
            self.deep_water = False

    def transform_periods(self, scalar):
        """ Transform the wave periods with the Rayleigh characteristics

        Transforms the missing wave periods in the LimitState with the
        characteristics for Rayleigh distributed waves in deep water,
        by using the wave periods which are in the LimitState.

        Table 2: characteristic wave periods (Van den Bos and Verhagen, 2018)

        +-------------------+--------------+
        | Name              |     T/Tp     |
        +===================+==============+
        | :math:`T_{p}`     |      1       |
        +-------------------+--------------+
        | :math:`T_{m}`     | 0.75 to 0.85 |
        +-------------------+--------------+
        | :math:`T_{s}`     | 0.90 to 0.95 |
        +-------------------+--------------+
        | :math:`T_{m-1.0}` |      1.1     |
        +-------------------+--------------+

        Parameters
        ----------
        scalar : float
            Should be between 0 and 1, 0 being the lower bound, 1 the
            upper bound of the characteristics. For instance, scalar of
            1 results in Tm/Tp = 0.85.

        Raises
        ------
        InputError
            If the scalar is not between 0 and 1
        """
        if scalar > 1 or scalar < 0:
            raise ValueError('The scalar should be between 0 and 1')

        if not self.deep_water:
            limitstate_warning(
                'Transforming wave periods is only valid in deep water')

        if 'Tp' not in self.conditions:
            if 'Tm' in self.conditions:
                Tm = self.conditions['Tm']
                self.conditions['Tp'] = Tm/(0.75 + scalar*0.1)
            elif 'T13' in self.conditions:
                T13 = self.conditions['T13']
                self.conditions['Tp'] = T13/(0.9 + scalar*0.05)
            elif 'T_m_min_1' in self.conditions:
                T_m_min_1 = self.conditions['T_m_min_1']
                self.conditions['Tp'] = T_m_min_1/1.1

        if 'Tp' in self.conditions:
            Tp = self.conditions['Tp']
            if 'Tm' not in self.conditions:
                self.conditions['Tm'] = (0.75 + scalar*0.1) * Tp
            if 'T13' not in self.conditions:
                self.conditions['T13'] = (0.9 + scalar*0.05) * Tp
            if 'T_m_min_1' not in self.conditions:
                self.conditions['T_m_min_1'] = 1.1*Tp

    def get_Hs(self, definition):
        """ Get the significant wave height

        Multiple definitions are used for the significant wave height,
        and formulas use different definitions. This method returns the
        desired definition of the significant wave height. For
        overtopping this is Hm0 and H1/3 is used in the Van der Meer
        and Hudson formula.

        .. note::
           If Hm0, Hs or H13 are in the LimitState this method will
           **always** return a significant wave height. The table
           depicts the order in which Hs is returned if a value is
           missing. For example: if the chosen definition is Hm0, then
           first Hm0 is returned, if Hm0 is not included in the
           Limitstate Hs will be retuned, if Hs is also missing H13
           will be returned.

           +------------+-------+-------+-------+
           | definition |  Hm0  |  Hs   |  H13  |
           +============+=======+=======+=======+
           | 1          |  Hm0  |  Hs   |  H13  |
           +------------+-------+-------+-------+
           | 2          |  Hs   |  H13  |  Hs   |
           +------------+-------+-------+-------+
           | 3          |  H13  |  Hm0  |  Hm0  |
           +------------+-------+-------+-------+

        Parameters
        ----------
        definition : {'Hm0', 'Hs', 'H13'}
            definition of the significant wave height to use

        Returns
        -------
        Hs : float
            significant wave height based on the definition

        Raises
        ------
        InputError
            If there is no significant wave height (Hm0, Hs, H1/3) in
            the LimitState
        KeyError
            If the definition is not 'Hm0', 'H13' or 'Hs'
        """
        info = None
        msg = 'No significant wave height in the LimitState'

        if definition == 'Hm0':
            if 'Hm0' in self.conditions:
                H = self.conditions['Hm0']
            elif 'Hs' in self.conditions:
                H = self.conditions['Hs']
                info = f'used Hs instead of {definition} from {self.label}'
            elif 'H13' in self.conditions:
                H = self.conditions['H13']
                info = f'used Hs instead of {definition} from {self.label}'
            else:
                raise InputError(msg)

        elif definition == 'H13':
            if 'H13' in self.conditions:
                H = self.conditions['H13']
            elif 'Hs' in self.conditions:
                H = self.conditions['Hs']
                info = f'used Hs instead of {definition} from {self.label}'
            elif 'Hm0' in self.conditions:
                H = self.conditions['Hm0']
                info = f'used Hm0 instead of {definition} from {self.label}'
            else:
                raise InputError(msg)

        elif definition == 'Hs':
            if 'Hs' in self.conditions:
                H = self.conditions['Hs']
            elif 'H13' in self.conditions:
                H = self.conditions['H13']
                info = f'used H13 instead of {definition} from {self.label}'
            elif 'Hm0' in self.conditions:
                H = self.conditions['Hm0']
                info = f'used Hm0 instead of {definition} from {self.label}'
            else:
                raise InputError(msg)

        else:
            msg = (f'{definition} not a valid key for get_Hs, use \'Hm0\''
                   ', \'H13\' or \'Hs\' instead')
            raise KeyError(msg)

        if info is not None:
            limitstate_warning(info)

        return H

    def L(self, period, deep_water=False):
        """ Compute the wave length with the dispersion relation

        .. math::
           L = L_{o} \\tanh \\left( \\frac{2 \\pi h}{L}\\right)

        in which the deep water wave length is:

        .. math::
           L_{o} = \\frac{g T^{2}}{2 \\pi}

        Parameters
        ----------
        period : str
            keyword argument of a wave period, must be in LimitState
        deep_water : bool, optional, default: False
            if False the dispersion relation will be used, if True
            the deep water wave length will be used. Note that this
            parameter is not equal to the attribute deep_water

        Raises
        ------
        KeyError
            If the specified wave period is not in the LimitState
        """
        T = self.conditions[period]

        if deep_water:
            # user wants deep water wave length
            wave_length = 9.81*T**2/(2*np.pi)
        else:
            # user wants to use the dispersion relation
            wave_length = dispersion(T=T, h=self.h)

        return wave_length

    def s(self, number=None, H=None, T=None):
        """ Compute the fictitious wave steepness

        .. math::
           s_{0} = \\frac{H}{L_{o}}

        the fictitious wave steepness combines the value of the wave
        height at the location of the breakwater with the deep water
        wave length. It is possible to select a number or specify which
        wave height and period must be used for the computation.

        .. note::
           When computing the fictitious wave steepness the significant
           wave height is needed, to get the significant wave height
           form the LimitState the method :py:meth:`get_Hs` is used.

        Parameters
        ----------
        number : {'mean', 'spectral'}, optional, default: None
            definition to use, will automatically select the correct
            definitions for H and T
        H : str, optional, default: None
            wave height in the LimitState
        T : str, optional, default: None
            wave period in the LimitState

        returns
        s_0 : float
            the fictitious wave steepness

        Raises
        ------
        InputError
            If there is no significant wave height (Hm0, Hs, H1/3) in
            the LimitState.
        KeyError
            if the specified wave period is not in the LimitState
        """
        if number == 'mean':
            H = self.get_Hs('H13')
            L = self.L(period='Tm', deep_water=True)
        elif number == 'spectral':
            H = self.get_Hs('Hm0')
            L = self.L(period='T_m_min_1', deep_water=True)
        else:
            H = self.conditions[H]
            L = self.L(period=T, deep_water=True)

        steepness = H/L

        return steepness

    def surf_similarity(
        self, alpha, number=None, H=None, T=None):
        """ Compute the surf similarity parameter

        .. math::
           \\xi = \\frac{tan{\\alpha}}{\\sqrt{s_{o}}}

        Computes the surf similarity parameter, also known as the
        Iribarren number. It is possible to select a number or specify
        which wave height and period must be used for the computation.

        Parameters
        ----------
        alpha : float
            slope of the structure [rad]
        number : {'mean', 'spectral'}, optional, default: None
            definition to use, will automatically select the correct
            definitions for H and T
        H : str, optional, default: None
            wave height in the LimitState
        T : str, optional, default: None
            wave period in the LimitState

        Returns
        -------
        xi : float
            the surf similarity parameter [-]

        Raises
        ------
        InputError
            If there is no significant wave height (Hm0, Hs, H1/3) in
            the LimitState
        KeyError
            If the specified wave period is not in the LimitState
        """
        if H is None and T is None:
            s = self.s(number=number)
            xi = np.tan(alpha)/np.sqrt(s)
        else:
            s = self.s(H=H, T=T)
            xi = np.tan(alpha)/np.sqrt(s)

        return xi

    def get_H2(self, slope_foreshore):
        """ Get the wave height exceeded by 2% of the waves, H2%

        Attempts to return H2% from the defined limit state, if not
        defined in the limit state it is computed with
        :py:class:`BattjesGroenendijk`.

        Parameters
        ----------
        slope_foreshore : float
            the slope of the foreshore [rad]

        Returns
        -------
        H2_per : float
            the wave height exceeded by 2% of the waves [m]

        Raises
        ------
        KeyError
            If Hm0 is not in the LimitState
        """
        if 'H2_per' in self.conditions:
            return self.conditions['H2_per']

        else:
            # get Hm0 and initiate battjes class
            Hm0 = self.conditions['Hm0']
            battjes = BattjesGroenendijk(
                Hm0=Hm0, h=self.h, slope_foreshore=slope_foreshore)

            # add H2% to the LimitState and return the value
            self.conditions['H2_per'] = battjes.get_Hp(0.02)
            return self.conditions['H2_per']

    def Sd(self, G, nv):
        """ Compute the damage level parameter Sd

        This method approximates Sd by using equation 5.150 from the
        Rock Manual (CIRIA, CUR, CETMEF, 2007), and adds Sd to the
        LimitState. The equation to compute Sd is given as:

        .. math::
           S_{d} = \\frac{N_{od}}{G (1 - n_{v})}

        Parameters
        ----------
        G : float
            gradation factor [-]
        nv : float
            porosity of the armour layer [-]

        Returns
        -------
        Sd : float
            damage level parameter [-]

        Raises
        ------
        KeyError
            If Nod is not in the LimitState, use :py:meth:`Nod` instead
        """
        if 'Sd' in self.conditions:
            print('WARNING: This method overwrites the existing Sd with the'
                  ' computed Sd')

        Sd = self.conditions['Nod']/(G*(1-nv))

        self.conditions['Sd'] = Sd

        return Sd

    def Nod(self, G, nv):
        """ Compute the damage number Nod

        This method approximates Nod by using equation 5.150 from the
        Rock Manual (CIRIA, CUR, CETMEF, 2007), and adds Nod to the
        LimitState. The equation to compute Nod is given as:

        .. math::
           N_{od} = G (1 - n_{v}) S_{d}

        Parameters
        ----------
        G : float
            gradation factor [-]
        nv : float
            porosity of the armour layer [-]

        Returns
        -------
        Nod : float
            damage number [-]

        Raises
        ------
        KeyError
            If Sd is not in the LimitState, use :py:meth:`Sd` instead
        """
        if 'Nod' in self.conditions:
            print('WARNING: This method overwrites the existing Nod with the'
                  ' computed Nod')

        Nod = G * (1-nv) * self.conditions['Sd']

        self.conditions['Nod'] = Nod

        return Nod



import context
import unittest
import numpy as np

from breakwater.rubble import RockRubbleMound, ConcreteRubbleMound
from breakwater.conditions import LimitState
from breakwater.material import RockGrading, Xbloc
from breakwater.utils.exceptions import no_warnings


class TestRockRubbleMound(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        no_warnings()

        ULS = LimitState(
            deep_water=False, h=10, Nod=0.5, label='ULS', Sd=2, q=10,
            Hm0=3.22, T_m_min_1=7.66, Tp=6.96, Tm=6.33)
        NEN = RockGrading()

        cls.bw = RockRubbleMound(
            slope=(3,4), slope_foreshore=(1,100), B=6.4, rho_w=1000,
            Dn50_core=0.3, LimitState=ULS, Grading=NEN, safety=1, id=1,
            N=3000)

    def test_structure_armour(self):
        armour = self.bw.structure['armour']
        computed = [1.399, 'HMA_6000/10000', 1.445, 0, 2]
        for i, (param, val) in enumerate(armour.items()):
            self.assertAlmostEqual(val, computed[i], 3)

    def test_structure_underlayer(self):
        underlayer = self.bw.structure['underlayer']
        computed = [[0.586, 0.671],
                    ['HMA_300/1000', 'HMA_1000/3000'],
                    [0.615, 0.895],
                    'see armour',
                    2]
        for i, (param, vals) in enumerate(underlayer.items()):
            if isinstance(vals, list):
                for j, val in enumerate(vals):
                    self.assertAlmostEqual(val, computed[i][j], 3)
            else:
                self.assertEqual(vals, computed[i])

    def test_structure_filter(self):
        filter = self.bw.structure['filter layer']
        computed = [[None, 0.306, 0.415],
                    [None, 'LMA_40/200', 'LMA_60/300'],
                    [None, 0.335, 0.388],
                    'see armour',
                    2]
        for i, (param, vals) in enumerate(filter.items()):
            if isinstance(vals, list):
                for j, val in enumerate(vals):
                    self.assertAlmostEqual(val, computed[i][j], 3)
            else:
                self.assertEqual(vals, computed[i])

    def test_structure_toe(self):
        toe = self.bw.structure['toe']
        computed = [0.790, 'HMA_1000/3000', 0.895, 0]
        for i, (param, val) in enumerate(toe.items()):
            self.assertAlmostEqual(val, computed[i], 3)

    def test_area(self):
        areas = self.bw.area('a')
        computed = [109.035, 52.574, 171.323, 9.613]
        for i, (layer, area) in enumerate(areas.items()):
            self.assertAlmostEqual(area, computed[i], 3)


class TestConcreteRubbleMound(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        no_warnings()

        ULS = LimitState(
            deep_water=False, h=10, Nod=0.5, q=10, Hm0=3.22, Tp=6.96, Tm=6.33,
            T_m_min_1=7.66, label='ULS')
        SLS = LimitState(
            deep_water=False, h=9, Nod=0.5, q=10, Hm0=2.52, Tp=6.96, Tm=6.33,
            T_m_min_1=7.66, label='ULS')
        NEN = RockGrading()
        xbloc = Xbloc()

        cls.bw = ConcreteRubbleMound(
            slope=(3,4), slope_foreshore=(1,100), B=6.4, rho_w=1000,
            Dn50_core=0.3, LimitState=[SLS, ULS], id=1, ArmourUnit=xbloc,
            Grading=NEN, safety=1)

    def test_structure_armour(self):
        armour = self.bw.structure['armour']
        computed = [0.949, 1, 1, 1, 1]
        for i, (param, val) in enumerate(armour.items()):
            self.assertAlmostEqual(val, computed[i], 3)

    def test_structure_underlayer(self):
        underlayer = self.bw.structure['underlayer']
        computed = [[0.392, 0.532],
                    ['LMA_60/300', 'HMA_300/1000'],
                    [0.388, 0.615],
                    'see armour',
                    2]
        for i, (param, vals) in enumerate(underlayer.items()):
            if isinstance(vals, list):
                for j, val in enumerate(vals):
                    self.assertAlmostEqual(val, computed[i][j], 3)
            else:
                self.assertEqual(vals, computed[i])

    def test_structure_filter(self):
        self.assertRaises(KeyError, lambda: self.bw.structure['filter layer'])

    def test_structure_toe(self):
        toe = self.bw.structure['toe']
        computed = [0.450, 'HMA_300/1000', 0.615, 1]
        for i, (param, val) in enumerate(toe.items()):
            self.assertAlmostEqual(val, computed[i], 3)


if __name__ == '__main__':
    unittest.main()



import os
import pandas as pd
import numpy as np
import scipy.stats as stats
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec

from ..utils.exceptions import user_warning

try:
    from mpl_toolkits.basemap import Basemap
except ImportError:
    user_warning('The Basemap package is required to use BreakwaterDatabase') 

class BreakwaterDatabase:
    """ Breakwater Database

    Import the breakwater database developed by (Allsop et al., 2009)
    consist of completed breakwater projects with data ranging from
    design wave height to contractor. The constructed breakwaters are
    classified by breakwater type, the following types are currently
    included: Rubble Mound, Composite, Berm, Caisson and Revetments.

    The breakwater database is a separate module of :py:obj:`breakwater`
    and can be imported with the following command:

    .. code-block:: python

       from breakwater.database import BreakwaterDatabase

    .. note::
       To use this module the :py:obj:`Basemap` package is required,
       this dependency is additional to the dependencies mentioned in
       Section `2.1 <install.html#Dependencies>`__. See the following
       `link <https://matplotlib.org/basemap/users/installing.html>`__
       for an installation of :py:obj:`Basemap`

    Parameters
    ----------
    update : bool, optional, default: False
        if False the data is not updated and the included data is
        loaded, if True the database is loaded from the :py:attr:`source`

    Attributes
    ----------
    df : pd.DataFrame
        DataFrame of the breakwater database
    source : str
        url of the source
    """

    def __init__(self, update=False):
        """ See help(BreakwaterDatabase) for more info """
        # set source of the database as attribute
        self.source = 'http://kennisbank-waterbouw.tudelft.nl/breakwaters/printall.php'

        # check if data must be updated from the source
        if update:
            # updata data from the source
            dataframes = pd.read_html(self.source)

            # get the correct df from the dfs
            for dataframe in dataframes:
                # check df of the database by checking the number of columns
                if dataframe.shape[1] > 10:
                    # correct df
                    df = dataframe
                    break

            # check if headers are correct
            if df.columns.dtype != object:
                # set first row as header
                header = df.iloc[0]
                df = df[1:]
                df.columns = header

            # set id coluimn as index
            df.set_index('id', inplace=True)
            df.index = df.index.astype(np.int64)

            # fix column names of front and rear
            names = ['armour', '(unit)', 'size', 'slope']
            new_columns = []

            # iterate over the columns
            for i, column in enumerate(df.columns):
                # only columns after front and rear must be changed
                if (column == 'front' or column == 'rear'
                        and pd.isnull(df.columns[i+1:i+4]).any()):
                    # change names of columns
                    for name in names:
                        new_columns.append(f'{column} {name}')

                elif pd.isnull(column):
                    pass

                else:
                    new_columns.append(column)

            # update column names
            df.columns = new_columns

            # convert numeric values to numeric values
            df = df.apply(pd.to_numeric, errors='ignore')

            # format types
            types_fmt = {
                'Rubble mound': 'Rubble Mound',
                'rubble mound': 'Rubble Mound',
                'Composite breakwater': 'Composite',
                'composite': 'Composite',
                'berm breakwater': 'Berm',
                'Berm breakwater': 'Berm',
                'Berm Breakwater': 'Berm',
                'Caisson breakwater': 'Caisson',
                'caisson breakwater': 'Caisson',
                'caisson': 'Caisson'}
            df.type = df.type.replace(types_fmt)

            # fix incorrect classified types
            units = ['Tetrapod', 'Antifer', 'Cubes', 'Accropode',
                     'Accropode II', 'Core-Loc', 'Xbloc', 'XblocPlus',
                     'Dolos', 'COB', 'Stabit', 'Cubipod']
            for unit in units:
                df.type = df.type.replace(unit, 'Rubble Mound')

        else:
            # load data from csv file
            file = os.path.dirname(os.path.abspath(__file__))
            df = pd.read_csv(f'{file}\\database.csv', index_col=0)

        # set df as attribute
        self.df = df

        # set markers and colors as private attributes
        self._markers = {
            'Rubble Mound': 'o', 'Caisson': 's', 'Berm': 'D',
            'Revetment': '>', 'Composite': '*', 'Unclassified': 'p'}
        self._colors = {
            'Rubble Mound': '#1f77b4', 'Caisson': '#ff7f0e', 'Berm': '#2ca02c',
            'Revetment': '#d62728', 'Composite': '#9467bd',
            'Unclassified': '#bcbd22'}

    @staticmethod
    def _del_zeros(df):
        """ Delete rows containing a zero """
        # make list to store ids to delete
        to_delete = []

        # iterate over the df
        for id, row in df.iterrows():
            # iterate over the row
            for param, value in row.iteritems():
                # check if value is zero
                if value == 0:
                    # row must be deleted
                    to_delete.append(id)
                    break

        # delete rows and return df
        return df.drop(labels=to_delete, axis=0)

    @staticmethod
    def _show_unclassified(show, df):
        """ Helper method to filter the df for unclassified bw types """
        # check if unclassified must be plotted
        if show:
            # replace nan with unclassified so label is correct in plot
            df.type = df.type.replace(np.nan, 'Unclassified')
        else:
            # remove unclassified breakwaters
            df.dropna(subset=['type'], inplace=True)

        return df

    @staticmethod
    def _validate_excludes(input):
        """ Verify the input for the correct type """
        # check input of excludes
        if input is not None and not isinstance(input, list):
            # raise TypeError for incorrect type of exclude
            raise TypeError(
                ('Types to exclude must be given in a list, not as '
                f'{type(input).__name__}'))

    @staticmethod
    def _validate_all_excludes(input):
        """ Validate if specified types have not been excluded """
        if input:
            # not all given types were in the database,
            # and were thus not excluded, therefore show warning
            not_excluded = ', '.join(input)
            user_warning(
                ('The following types in exclude are not in the database: '
                f'{not_excluded}'))

    @property
    def unclassified(self):
        """ Get the number of unclassified breakwaters """
        # get the unclassified breakwaters
        return self.df.type.isna().sum()

    def report(self, save=True, save_path='data_report.xlsx', decimals=3):
        """ Make a report of the data in the database

        Method to generate a data report of the database. For each
        breakwater type the total and missing number of datapoints is
        determined. Furthermore, for numerical values the mean and
        standard deviation is also computed.

        .. note:
           Note that the mean and standard deviation is not computed
           for the following parameters: start, finish, front size,
           rear size, latitude and longitude

        Parameters
        ----------
        save : bool, optional, default: True
            If True an Excel version of the data report is generated,
            use :py:obj:`save_path` to specify a save path
        save_path : str
            File path to save the Excel file to
        decimals : int, optional, default: 3
            Number of decimal places to round to (default: 0). If
            decimals is negative, it specifies the number of
            positions to the left of the decimal point.

        Returns
        -------
        pd.DataFrame
            if :py:obj:`save` is False a DataFrame of the report is
            returned
        """
        # create df to store data report in
        data_report = pd.DataFrame()

        # drop unclassified breakwaters
        df = self._show_unclassified(False, self.df)

        # define columns of which the mean and standard deviation does have
        # to be computed
        no_computations_required = [
            'start', 'finish', 'front size', 'rear size', 'Lat', 'Lon']

        # make list for headers and subheaders
        headers, subheaders = ['Type'], ['']

        # set bool to track if subheaders have been added for wall column
        add_subheaders_wall = True

        # iterate over the bw types
        for i, bw_type in enumerate(df.type.unique()):
            # get only the bws of the current type
            bws = df[df.type == bw_type]

            # create dict with bw_type to store data
            type_data = {'type': [bw_type]}

            # iterate over the columns with dtype (for computing mean, etc)
            for column, dtype in bws.dtypes.iteritems():
                # add name of the column as a header if first iteration
                if i == 0 and column != 'type':
                    headers.append(column)

                # check if column has the bw_type
                if column == 'type':
                    # pass as this has already been added
                    continue

                # check type of the column
                elif dtype == np.object:
                    # column contains strings or is a coordinate
                    # count total number of datapoints
                    total = len(bws[column].values)

                    # check if column contains the slope
                    if 'slope' in column:
                        # check bw type
                        if bw_type == 'Caisson' or bw_type == 'Composite':
                            # have slope 1:0 but this is not needed since
                            # these are vertical structures
                            type_data[column] = ['-']

                            # add subheader if first iteration
                            if i == 0:
                                subheaders.append('')
                            continue

                        else:
                            # slope 1:0 is specified when the slope unknown
                            # so replace these values with nan
                            column_data = bws[column].replace('1:0', np.nan)

                            missing = column_data.isna().sum()

                    else:
                        # count missing datapoints
                        missing = bws[column].isna().sum()

                    # add to column
                    type_data[column] = [f'{total-missing}/{total}']

                    # add subheader if first iteration
                    if i == 0:
                        subheaders.append('')

                elif column in no_computations_required:
                    # replace zeros for nan
                    column_data = bws[column].replace(0, np.nan)

                    # compute total and missing datapoints
                    total = len(bws[column].values)
                    missing = column_data.isna().sum()

                    # add to column
                    type_data[column] = [f'{total-missing}/{total}']

                    # add subheader if first iteration
                    if i == 0:
                        subheaders.append('')

                else:
                    # column is an int or float
                    # check if bw type is not caisson and column is wall
                    if column == 'wall' and bw_type != 'Caisson':
                        # wall is not a parameter of other structures
                        type_data[f'{column} no'] = ['-']
                        type_data[f'{column} comp'] = ['-']

                        # add subheader if first iteration
                        if add_subheaders_wall:
                            headers.append(column)
                            subheaders.append('')
                            subheaders.append('')
                            add_subheaders_wall = False
                        continue

                    # replace zeros for nan
                    column_data = bws[column].replace(0, np.nan)

                    # compute total and missing datapoints
                    total = len(bws[column].values)
                    missing = column_data.isna().sum()

                    # compute average and standard deviation
                    mean = column_data.mean()
                    std = column_data.std()

                    # add data to dict
                    type_data[f'{column} no'] = [f'{total-missing}/{total}']
                    type_data[f'{column} comp'] = [f'{mean}±{std}']

                    # append extra column to headers and set subheaders
                    if i == 0:
                        headers.append(column)
                        subheaders.append('datapoints')
                        subheaders.append('μ ± σ')

                        # check if column is wall for setting the bool
                        if column == 'wall':
                            # set to False because has been added here
                            add_subheaders_wall = False

            # add dict to df
            temp_df = pd.DataFrame(data=type_data)
            data_report = data_report.append(
                temp_df, ignore_index=True, sort=False)

        # update columns of the df with double row of headers
        data_report.columns = [headers, subheaders]

        # check if df must be saved
        if save:
            # save df
            data_report.to_excel(save_path)
        else:
            # return df
            return data_report

    def correlation(self, param1, param2, bw_type=None, method='pearson'):
        """ Compute the correlation between two parameters

        Parameters
        ----------
        param1 : str
            name of parameter 1
        param2 : str
            name of parameter 2
        bw_type : str, optional, default: None
            if specified only the values of the given bw_type are
            considered
        method : {pearson, spearman}, optional, default: pearson
            method of correlation

        Returns
        -------
        tuple
            correlation between the two parameters and the p-value
        """
        # check if bw_type is given
        if bw_type is not None:
            # only for bw_type
            pass

        else:
            # for all bw types
            # get the two columns and remove zeros
            filtered_df = self._del_zeros(self.df[[param1, param2]])

        # method of correlation
        if method == 'pearson':
            # pearson method
            corr = stats.pearsonr(
                filtered_df[param1].values, filtered_df[param2].values)

        elif method == 'spearman':
            # spearman method
            corr = stats.spearmanr(
                filtered_df[param1].values, filtered_df[param2].values)

        else:
            raise NotImplementedError(
                f'{method} is not supported, must be pearson or spearman')

        return corr

    def cross_section(self, id, B=None, Rc=None, h=None, slope=None):
        """ Plot a cross section of the breakwater

        Method to plot a cross section of a breakwater in the database.
        The breakwater is selected by the id of the breakwater. In case
        data is missing to plot the breakwater it is possible to specify
        these as arguments.

        .. warning::
           plot function currently only supports Rubble Mound and
           caisson breakwaters

        Parameters
        ----------
        id : int
            id of the breakwater to plot
        B : float, optional, default: None
            specify custom crest width
        Rc : float, optional, default: None
            specify custom crest height
        h : float, optional, default: None
            specify a custom water level
        slope : tuple, optional, default: None
            specify a custom slope, must be specified as a tuple (V, H)

        Raises
        ------
        ValueError
            If data required to plot a cross section is missing
        """
        # set custom slope if not specified for protecting _validate
        if slope is None:
            slope = (0,0)

        # get the data of the bw
        bw = self.df[self.df.index == id]
        type = bw.type.values[0]

        # get hydraulic data
        depth = _validate_plot_vals('depth', bw['depth'].values[0], h)

        # get geometric data
        freeboard = _validate_plot_vals('Rc', bw['heigth'].values[0], Rc)
        width = _validate_plot_vals('B', bw['width'].values[0], B)

        # set infobox with general info
        cost = bw['cost(M$)'].values[0]
        infobox = '\n'.join((
            f'constructed between {bw.start.values[0]} and {bw.finish.values[0]}',
            f'cost = {cost} M$',
            f'owner = {bw.owner.values[0]}',
            f'contractor = {bw.contractor.values[0]}',
            f'consultant = {bw.consultant.values[0]}',
            f'Hs = {bw.Hs.values[0]} m',
            f'Tz = {bw.Tz.values[0]} s',
            f'Tp = {bw.Tp.values[0]} s',
            f'Rc = {freeboard} m',
            f'h = {depth} m',))

        # check type
        if type == 'Rubble Mound':
            # get the slope of the bw
            slope_databasse = bw['front slope'].values[0].split(':')
            V = _validate_plot_vals(
                'slope', float(slope_databasse[0]), slope[0])
            H = _validate_plot_vals(
                'slope', float(slope_databasse[1]), slope[1])

            # compute coordinates
            x1 = H*(depth+freeboard)/V
            xwlev = H*freeboard/V

            x = [-x1-0.5*width, -0.5*width, 0.5*width, x1+0.5*width]
            y = [0, depth+freeboard, depth+freeboard, 0]

            # set additional width for xmin and xmax
            increase_x = 0

            # add extra info of rubble mound
            armour = bw['front armour'].values[0]
            size = bw['front size'].values[0]
            unit = bw['front (unit)'].values[0]

            infobox = '\n'.join((
                infobox,
                f'armour = {armour} of {size} {unit}',
                f'slope = {V}:{H}'))

        elif type == 'Caisson':
            # compute coordinates
            xwlev = 0

            x = [-0.5*width, -0.5*width, 0.5*width, 0.5*width]
            y = [0, depth+freeboard, depth+freeboard, 0]

            # set additional width for xmin and xmax
            increase_x = 2*width

        else:
            raise NotImplementedError(
                f'{type} is currently not supported for plotting')

        # create figure
        fig, ax = plt.subplots(figsize=(10,5))

        # plot bw
        ax.plot(x,y, color='k', lw=1)

        # get the xmin and xmax
        xmin = ax.get_xlim()[0]*1.2 - increase_x
        xmax = ax.get_xlim()[1]*1.2 + increase_x

        # plot bottom and wlev (left and right)
        ax.axhline(y=0, color='peru', lw=1, zorder=5)
        ax.hlines(
            y=depth, xmin=xmin, xmax=-xwlev-0.5*width, color='dodgerblue', lw=1)
        ax.hlines(
            y=depth, xmin=xwlev+0.5*width, xmax=xmax, color='dodgerblue', lw=1)

        # place the info box with all info
        props = dict(boxstyle='round', facecolor='whitesmoke', alpha=0.5)
        ax.text(
            1.02, 0.99, infobox, transform=ax.transAxes, fontsize=9,
            verticalalignment='top', bbox=props)


        # format the figure
        ax.set_xlim(xmin, xmax)
        plt.title(
            (f'Cross section of the {type.lower()} breakwater at '
             f'{bw.harbour.values[0]}, {bw.country.values[0]}'))

        ax.set_aspect('equal', adjustable='box')
        ax.grid()
        fig.tight_layout()
        plt.show()

    def map(
            self, area=[], resolution='c', show_unclassified=False,
            exclude=None):
        """ Plot the breakwaters on a world map

        Method to plot all breakwaters with coordinates on a map of
        the world, or part of the world if an area is specified. Method
        uses :py:obj:`Basemap` to generate the map.

        Parameters
        ----------
        area : list, optional, default: []
            specify the coordinates of the area to plot. Use following
            format [llcrnrlon, llcrnrlat, urcrnrlon, urcrnrlat]
        resolution : str, optional, default: c
            resolution of the map to use. Can be c (crude), l (low),
            i (intermediate), h (high), f (full).
        show_unclassified : bool, optional, default: False
            True is unclassified breakwaters with a coordinate must be
            plotted, False is unclassified breakwaters must not be
            plotted.
        exclude : list, optional, default: None
            list of breakwater types to exclude from the plot
        """
        # validate exclude input for type
        self._validate_excludes(exclude)

        # create the figure
        plt.figure(figsize=(16,12))

        # check if an area is specified
        if not area:
            # plot full map
            m = Basemap(
                lat_0=0, lon_0=0, projection='robin', resolution=resolution)
            size = 7

        else:
            # plot the specified area
            m = Basemap(
                llcrnrlon=area[0], llcrnrlat=area[1], urcrnrlon=area[2],
                urcrnrlat=area[3], resolution=resolution)
            size = 15

        # edit lay-out of the map
        m.drawcountries(color='#d7d7d7')
        m.drawmapboundary(fill_color='#D0CFD4', linewidth=0)
        m.fillcontinents(color='#EFEFEF', lake_color='#D0CFD4')

        # remove bw's without a Lon and\or Lat
        filtered_df = self._del_zeros(self.df[['type', 'Lon', 'Lat']])

        # filter df for showing unclassified bw types
        filtered_df = self._show_unclassified(show_unclassified, filtered_df)

        # iterate over the bw types to plot them
        for i, bw_type in enumerate(filtered_df.type.unique()):
            # check if type is in exclude
            if exclude is not None and bw_type in exclude:
                # pass and delete type from exclude
                exclude.remove(bw_type)

            else:
                # get only the bws of the current type
                bws = filtered_df[filtered_df.type == bw_type]

                # plot the bws
                xpt, ypt = m(bws.Lon.values, bws.Lat.values)
                m.scatter(
                    xpt, ypt, s=size, alpha=1, label=bw_type,
                    c=self._colors.get(bw_type, '#17becf'),
                    marker=self._markers.get(bw_type, 'H'), zorder=2)

        # check if all types in exclude are in the database
        self._validate_all_excludes(exclude)

        # add legend, set tight_layout and show plot
        plt.legend(loc=1)
        plt.tight_layout()
        plt.show()

    def scatter(
            self, param1, param2, show_unclassified=False, exclude=None,
            min_data=5, xmax=None, ymax=None, bins_param1=10, bins_param2=10):
        """ Make a scatter plot of two parameters

        Method to generate a scatter plot with histograms for two
        parameters in the database.

        Parameters
        ----------
        param1 : str
            name of parameter 1
        param2 : str
            name of parameter 2
        show_unclassified : bool, optional, default: False
            True is unclassified breakwaters with a coordinate must be
            plotted, False is unclassified breakwaters must not be
            plotted.
        exclude : list, optional, default: None
            list of breakwater types to exclude from the plot
        min_data : int, optional, default: 5
            minimum number of datapoints required for plotting, if the
            data for a bw type is less than this limit it will be skipped
        xmax : float, optional, default: None
            maximum x coordinate of the scatter plot, by default this
            limit is automatically determined
        ymax : float, optional, default: None
            maximum y coordinate of the scatter plot, by default this
            limit is automatically determined
        bins_param1 : str
            number of bins for param2
        bins_param2 : str
            number of bins for param2
        """
        # validate exclude input for type
        self._validate_excludes(exclude)

        # filter the df to remove zeros
        filtered_df = self._del_zeros(self.df[['type', param1, param2]])

        # filter df for showing unclassified bw types
        filtered_df = self._show_unclassified(show_unclassified, filtered_df)

        # create the figure
        fig = plt.figure(figsize=(12,9))
        gs = GridSpec(4,4)

        # create the scatter plot
        scatter_plot = fig.add_subplot(gs[1:4,0:3])

        # make the histograms of the two parameters
        top_hist = fig.add_subplot(gs[0,0:3])
        right_hist = fig.add_subplot(gs[1:4,3])

        # create empty list for storing data of bw types
        hist_param1, hist_param2 = [], []

        # create empty list to store colors for histogram
        colors = []

        # iterate over the bw types
        for bw_type in filtered_df.type.unique():
            # check if type is in exclude
            if exclude is not None and bw_type in exclude:
                # pass and delete type from exclude
                exclude.remove(bw_type)

            else:
                # get only the bws of the current type
                bws = filtered_df[filtered_df.type == bw_type]

                # check if there is enough data
                if (len(bws[param1].values) <= min_data
                        or len(bws[param2].values) <= min_data):
                    # not enough data, show warning
                    user_warning(
                        f'{bw_type} was skipped because of a lack of data')

                else:
                    # enough data, plot on scatter
                    scatter_plot.scatter(
                        bws[param1].values, bws[param2].values, s=18,
                        c=self._colors.get(bw_type, '#17becf'), label=bw_type,
                        marker=self._markers.get(bw_type, 'H'), zorder=2)

                    # add used color to the list
                    colors.append(self._colors.get(bw_type, '#17becf'))

                    # add to hist lists
                    hist_param1.append(bws[param1].values)
                    hist_param2.append(bws[param2].values)

        # check if all types in exclude are in the database
        self._validate_all_excludes(exclude)

        # determine x and y limits, if not specified as arguments
        if xmax is None:
            # determine xmax
            xmax = np.round(scatter_plot.get_xlim()[1])

        if ymax is None:
            # determine ymax
            ymax = np.round(scatter_plot.get_ylim()[1])

        # generate the bins
        param1_bins = np.linspace(0, xmax, bins_param1)
        param2_bins = np.linspace(0, ymax, bins_param2)

        # add the histograms to the plot
        top_hist.hist(hist_param1, bins=param1_bins, color=colors)
        right_hist.hist(
            hist_param2, bins=bins_param2, orientation='horizontal',
            color=colors)

        # set x and y lims
        scatter_plot.set_xlim(0, xmax)
        scatter_plot.set_ylim(0, ymax)
        top_hist.set_xlim(0, xmax)
        right_hist.set_ylim(0, ymax)

        # add grid to all plots
        scatter_plot.grid()
        top_hist.grid()
        right_hist.grid()

        # set labels
        scatter_plot.set_xlabel(param1.capitalize())
        scatter_plot.set_ylabel(param2.capitalize())
        top_hist.set_ylabel('Frequency')
        right_hist.set_xlabel('Frequency')

        # remove ticks from histograms
        plt.setp(top_hist.get_xticklabels(), visible=False)
        plt.setp(right_hist.get_yticklabels(), visible=False)

        # add legend, set tight_layout and show plot
        scatter_plot.legend()
        plt.tight_layout()
        plt.show()

    def hist(
            self, param, show_unclassified=False, exclude=None, min_data=5,
            xmax=None, bins=10):
        """ Plot a histogram of a parameter

        Parameters
        ----------
        param : str
            name of the parameter
        show_unclassified : bool, optional, default: False
            True is unclassified breakwaters with a coordinate must be
            plotted, False is unclassified breakwaters must not be
            plotted.
        exclude : list, optional, default: None
            list of breakwater types to exclude from the plot
        min_data : int, optional, default: 5
            minimum number of datapoints required for plotting, if the
            data for a bw type is less than this limit it will be skipped
        xmax : float, optional, default: None
            maximum x coordinate of the histogram, by default this
            limit is automatically determined
        bins : int, optional, default: 10
            number of bins
        """
        # validate exclude input for type
        self._validate_excludes(exclude)

        # remove the zero values from the df
        df = self.df[['type', param]]
        df = df[df[param] != 0]

        # filter df for showing unclassified bw types
        filtered_df = self._show_unclassified(show_unclassified, df)

        # create lists to store labels, data and colors
        labels, data, colors = [], [], []

        # set variable to track maximum x
        xmax_computed = 0

        # iterate over the types in the df to get the data
        for bw_type in filtered_df.type.unique():
            # check if type is in exclude
            if exclude is not None and bw_type in exclude:
                # pass and delete type from exclude
                exclude.remove(bw_type)

            else:
                # get only the bws of the current type
                bws = filtered_df[filtered_df.type == bw_type]

                # check if there is enough data
                if len(bws[param].values) <= min_data:
                    # not enough data, show warning
                    user_warning(
                        f'{bw_type} was skipped because of a lack of data')

                else:
                    # enough data, plot on scatter
                    # add used color, data and label to the list
                    labels.append(bw_type)
                    data.append(bws[param].values)
                    colors.append(self._colors.get(bw_type, '#17becf'))

                    # get maximum value
                    if np.max(bws[param].values) >= xmax_computed:
                        xmax_computed = np.max(bws[param].values)

        # check if all types in exclude are in the database
        self._validate_all_excludes(exclude)

        # determine x limit, if not specified as arguments
        if xmax is None:
            # determine xmax
            xmax = np.round(xmax_computed)

        # generate the bins
        generated_bins = np.linspace(0, xmax, bins)

        # plot histogram
        plt.hist(data, bins=generated_bins, label=labels, color=colors)

        # format axis
        plt.xlim(0, xmax)
        plt.xlabel(f'{param}')
        plt.ylabel('Frequency')

        # set other lay-out
        plt.title(f'Frequency Histogram of {param}')
        plt.grid()
        plt.legend()
        plt.tight_layout()
        plt.show()


def _validate_plot_vals(param, val, given_val):
    """ Validate the values to plot """
    if val == 0:
        # check if val is given
        if given_val != 0:
            # return specified value
            return given_val
        else:
            # raise error
            raise ValueError(
                (f'No value for {param} in the database, use arguments to '
                  'specify custom value'))
    else:
        # return value
        return val



import os
import sys
import string
try:
    import xlsxwriter
except ImportError:
    pass

from ._kwarg_validator import _RM_vkwargs, _C_vkwargs
from .exceptions import InputError, NotSupportedError

def _parameters(workbook, structure):
    """ Make sheet with input for parameters """
    # set first row and col
    row = 0
    first_col = 0

    # parameters to skip as these are object and not parameters
    skip = ['LimitState', 'Grading', 'ArmourUnit', 'BermMaterial']

    # make sheet
    sheet1 = workbook.add_worksheet(name='Parameters')

    # get parameters for the given structure
    params = {}
    if 'RRM' in structure and 'CRM' in structure:
        params.update(_RM_vkwargs(type='both'))
    elif 'RRM' in structure:
        params.update(_RM_vkwargs(type='Rock'))
    elif 'CRM' in structure:
        params.update(_RM_vkwargs(type='ArmourUnit'))

    if 'RC' in structure and 'CC' in structure:
        params.update(_C_vkwargs(type='both'))
    elif 'RC' in structure:
        params.update(_C_vkwargs(type='Rock'))
    elif 'CC' in structure:
        params.update(_C_vkwargs(type='ArmourUnit'))

    # set format to use
    fmt_header = workbook.add_format(
        {'bold': True, 'locked': True, 'right': 1, 'bottom': 2,
         'valign': 'vcenter'})

    fmt_header_center = workbook.add_format(
        {'bold': True, 'align': 'center', 'locked': True, 'right': 1,
         'bottom': 2, 'valign': 'vcenter'})

    fmt_header_center_var = workbook.add_format(
        {'bold': True, 'align': 'center', 'locked': True, 'right': 1,
         'valign': 'vcenter'})

    column = workbook.add_format({'right': 1, 'locked': True})

    column_center = workbook.add_format(
        {'align': 'center', 'locked': False, 'right': 1})
    required_column  = workbook.add_format(
        {'bg_color': '#FFC7CE', 'font_color': '#9C0006', 'align': 'center',
         'locked': False, 'right': 1})

    non_varying = workbook.add_format(
        {'align': 'center', 'bg_color': '#A6A6A6', 'locked': True, 'right': 1})

    sheet1.set_column(first_col, first_col, 20)
    sheet1.set_column(first_col+1, first_col+2, 10)

    # Turn worksheet protection on.
    options = {
        'format_cells':          True,
        'select_locked_cells':   False,
    }
    sheet1.protect('', options)

    # set the header of the columns
    headers = ['Parameter', 'Value', 'Varying']
    fmt = [fmt_header, fmt_header_center, fmt_header_center_var]
    for col, header in enumerate(headers):
        letter = string.ascii_uppercase[first_col+col]
        if header == 'Varying':
            # horizontal merge
            letter2 = string.ascii_uppercase[first_col+col+2]
            merge_range = f'{letter}{row+1}:{letter2}{row+1}'
            index_varying = col
        else:
            # vertical merge
            merge_range = f'{letter}{row+1}:{letter}{row+2}'

        sheet1.merge_range(merge_range, header, fmt[col])

    # increment row for writing the next row
    row += 1

    # write subheaders of Varying
    sheet1.write_string(row, index_varying, 'Min', fmt_header_center)
    sheet1.write_string(row, index_varying+1, 'Max', fmt_header_center)
    sheet1.write_string(row, index_varying+2, 'Num', fmt_header_center)

    # add parameters to sheet
    row += 1
    for key, info in params.items():
        if key not in skip:
            # write parameter to the 1st column
            sheet1.write_string(row, first_col, key, column)

            # write value to 2nd column if not required
            if not info['Required']:
                # not a required parameter, so add default value
                default = info['Default']

                # check if default is a tuple since these are not supported
                if isinstance(default, tuple) or isinstance(default, list):
                    # convert to string
                    default = str(default)

                # write to cel
                sheet1.write(row, first_col+1, default, column_center)
            else:
                # required parameter, thus write empty
                sheet1.write_blank(row, first_col+1, None, column_center)

            # write varying parameters to third column
            form_constant = f'ISBLANK(B{row+1})'

            if not info['Constant']:
                # parameter is allowed to vary
                varying_format = column_center

                # check if requried to set the correct formula
                if info['Required']:
                    # add conditional format to required values allowed to vary
                    form_varying = f'SUMPRODUCT(--(C{row+1}:E{row+1}<>""))<3'
                    form = (f'IF(({form_varying})=TRUE, IF({form_constant}='
                             'TRUE, TRUE, FALSE), FALSE)')
                    sheet1.conditional_format(
                        row, first_col+1, row, first_col+1,
                        {'type': 'formula', 'criteria': form,
                            'format': required_column})

                # add formula for data validation
                form_validation = (f'IF((SUMPRODUCT(--(B{row+1}:E{row+1}<>"")'
                                    ')=4)=TRUE, "Parameter set as constant '
                                    'and varying parameter, please choose one'
                                    '.", "")')
                sheet1.write_formula(row, first_col+5, form_validation)

            else:
                # parameter is not allowed to vary
                varying_format = non_varying

                if info['Required']:
                    sheet1.conditional_format(
                        row, first_col+1, row, first_col+1,
                        {'type': 'formula', 'criteria': form_constant,
                        'format': required_column})

            sheet1.write_blank(row, index_varying, None, varying_format)
            sheet1.write_blank(row, index_varying+1, None, varying_format)
            sheet1.write_blank(row, index_varying+2, None, varying_format)

            # increment row
            row += 1

def _sheet_generator(workbook, name, headers, required=None, num_rows=25):
    """ Make sheet for with input

    Make input sheet for the LimitState, RockGrading or ArmourUnits
    """
    # set number of columns and start row
    num_cols = len(headers)
    first_col = 0
    row = 0

    # add sheet to workbook
    sheet1 = workbook.add_worksheet(name)

    # set cel formatting
    # headers
    fmt_header = workbook.add_format(
        {'bold': True, 'locked': True, 'right': 1, 'bottom': 2})

    fmt_centered_header = workbook.add_format(
        {'bold': True, 'align': 'center', 'locked': True, 'bottom': 2,
         'right': 1})

    fmt_centered_header_last = workbook.add_format(
        {'bold': True, 'align': 'center', 'locked': True, 'bottom': 2})

    # first column
    fmt_param = workbook.add_format(
        {'right': 1, 'locked': False, 'bottom': 1})

    fmt_param_locked = workbook.add_format(
        {'right': 1, 'locked': True, 'bottom': 1})

    fmt_param_last_row = workbook.add_format({'right': 1, 'locked': False})

    # other columns
    fmt_val = workbook.add_format(
        {'align': 'center', 'locked': False, 'bottom': 1, 'right': 1})

    fmt_val_last_col = workbook.add_format(
        {'align': 'center', 'locked': False, 'bottom': 1})

    fmt_val_last_row = workbook.add_format(
        {'align': 'center', 'locked': False, 'right': 1})

    fmt_val_last_row_col = workbook.add_format(
        {'align': 'center', 'locked': False})

    fmt_empty = workbook.add_format({'locked': False})

    # turn on worksheet protection
    options = {
        'format_cells':          True,
        'insert_rows':           True,
        'delete_rows':           True,
        'select_locked_cells':   True,
    }
    sheet1.protect('', options)

    # set width of the sheets and write headers
    for col, header in enumerate(headers):
        if col == 0:
            # first col has a different width and format
            sheet1.set_column(col+first_col, col+first_col, 20)
            sheet1.write_string(row, col+first_col, header, fmt_header)
        elif col == num_cols-1:
            # format last column
            sheet1.set_column(col+first_col, col+first_col, 11)
            sheet1.write_string(
                row, col+first_col, header, fmt_centered_header_last)
        else:
            sheet1.set_column(col+first_col, col+first_col, 11)
            sheet1.write_string(
                row, col+first_col, header, fmt_centered_header)

    row += 1

    # check if there are required parameters
    if required is not None:
        # set the required required parameter
        param_index = headers.index('Parameter')
        value_index = headers.index('Value')
        for param in required:
            sheet1.write_string(
                row, param_index+first_col, param, fmt_param_locked)
            sheet1.write_blank(
                row, value_index+first_col, None, fmt_val_last_col)

            # increment row
            row += 1

    # add empty rows for input
    for i in range(num_rows):
        # check if last row
        if i < num_rows-1:
            # not the last row thus regular format
            for col in range(num_cols):
                if col == 0:
                    # different format for first column
                    sheet1.write_blank(row, col+first_col, None, fmt_param)
                elif col == num_cols-1:
                    # >2 columns thus different format for last column
                    sheet1.write_blank(
                        row, col+first_col, None, fmt_val_last_col)
                else:
                    sheet1.write_blank(row, col+first_col, None, fmt_val)
        else:
            # different format for last row
            for col in range(num_cols):
                if col == 0:
                    # different format for first column
                    sheet1.write_blank(
                        row, col+first_col, None, fmt_param_last_row)
                elif col == num_cols-1:
                    # >2 columns thus different format for last column
                    sheet1.write_blank(
                        row, col+first_col, None, fmt_val_last_row_col)
                else:
                    sheet1.write_blank(
                        row, col+first_col, None, fmt_val_last_row)

        # increment row
        row += 1

def generate_excel(filepath, input='configurations', structure=None):
    """ Generate excel file for design input

    Parameters
    ----------
    filepath : str
        location to save the excel input file
    input : str, optional, default: configurations
        specify which type of input excel must be generated, possible
        arguments: configurations (default), parameters, LimitState,
        Grading, ArmourUnits
    structure : {RRM, CRM, RC, CC}, optional, default: None
        structure for which the input sheet must be generated. RRM
        for a rubble mound with rock as armour layer, CRM for a rubble
        mound with concrete armour units as armour layer, RC for a
        vertical (composite) breakwater with rock as armour layer for
        the foundation and CC for a vertical (composite) breakwater with
        concrete armour units as armour layer for the foundation.

    Raises
    ------
    TypeError
        if the extension of the filepath is not .xlsx
    """
    # check if xlsxwriter is imported
    if 'xlsxwriter' not in sys.modules:
        # module was not imported, raise error
        raise ModuleNotFoundError(
            'The module Xlsxwriter is a required dependency for this function')

    # check if input is in supported input
    supported_input = [
        'configurations', 'parameters', 'LimitState', 'Grading',
        'ArmourUnits']
    if input not in supported_input:
        possible = ', '.join(supported_input)
        raise NotSupportedError(
            f'Excel input for {input} cannot be generated, must be {possible}')

    # get the extension of the file
    extension = os.path.splitext(filepath)[1]

    # check if an extension is included
    if not extension:
        # if not add .xlsx to the filepath
        filepath = f'{filepath}.xlsx'
    else:
        # there is an extension
        if extension != '.xlsx':
            # invalid extension
            raise TypeError(f'Extension {extension} is invalid, must be .xlsx')
        else:
            pass

    # generate the workbook
    workbook = xlsxwriter.Workbook(filepath)

    # add sheets
    if 'parameters' in input or 'configurations' in input:
        # get specified structure input
        if structure is not None:
            # convert the input of structure to a list
            if isinstance(structure, list):
                # must be a list so no change
                structure = structure
            elif isinstance(structure, str):
                # convert single input to list
                structure = [structure]
        else:
            raise InputError(
                ('Argument: structure must be specified when input is '
                 'parameters or configurations'))

        _parameters(workbook, structure)

    if 'LimitState' in input or 'configurations' in input:
        _sheet_generator(
            workbook, name='LimitState', headers=['Parameter', 'Value'],
            required=['h', 'label'])

    if 'Grading' in input or 'configurations' in input:
        _sheet_generator(
            workbook, name='RockGrading',
            headers=['Rock Class', 'M50 Lower', 'M50 Upper', 'NLL', 'NUL'])

    if 'ArmourUnits' in input or 'configurations' in input:
        # check if structure is given
        if structure is not None:
            # determine name of the sheet
            if 'CRM' in structure and 'CC' in structure:
                names = ['ArmourUnit', 'BermMaterial']
            elif 'CRM' in structure:
                names = ['ArmourUnit']
            elif 'CC' in structure:
                names = ['BermMaterial']
            else:
                names = []
        else:
            names = ['ArmourUnit']

        for name in names:
            _sheet_generator(
                workbook, name=name, headers=['Volume', 'D', 'h', 'Vc'])

    # save and close workbook
    workbook.close()



import numpy as np

from ..material import RockGrading, ConcreteArmour
from ..conditions import LimitState
from ..core.soil import Soil
from .exceptions import InputError

def _RM_vkwargs(type):
    """ Valid kwargs table for a rubble mound breakwater

    Construct a dictionary with all the valid kwargs for a rubble mound
    breakwater.

    Parameters
    ----------
    type : {'Rock', 'ArmourUnit', 'both'}
        type of rubble mound breakwater, R for rock and C for concrete
        armour units.

    Returns
    -------
    vkwargs : dict
        dictionary with all the valid kwargs for a rubble mound
        breakwater
    """
    # vkwargs for both types of rubble mound breakwaters
    vkwargs = {
        'slope': {'Default': None,
                  'Validator': lambda value: isinstance(value, tuple)
                                             and len(value) == 2,
                  'Correct': 'tuple with length 2',
                  'Constant': False,
                  'Required': True,
                  'unit': '[-]'},

        'slope_foreshore': {'Default': None,
                            'Validator': lambda value: isinstance(value, tuple)
                                                       and len(value) == 2,
                            'Correct': 'tuple with length 2',
                            'Constant': True,
                            'Required': True,
                            'unit': '[-]'},

        'rho_w' : {'Default': None,
                   'Validator': lambda value: isinstance(value, float)
                                              or isinstance(value, int),
                   'Correct': 'int or float',
                   'Constant': True,
                   'Required': True,
                   'unit': '[kg/m³]'},

        'B' : {'Default': None,
               'Validator': lambda value: isinstance(value, float)
                                          or isinstance(value, int),
               'Correct': 'int or float',
               'Constant': False,
               'Required': True,
               'unit': '[m]'},

        'LimitState' : {'Default': None,
                        'Validator': lambda value: isinstance(value, LimitState),
                        'Correct': 'LimitState',
                        'Constant': True,
                        'Required': True,
                        'unit': '[-]'},

        'Grading' : {'Default': None,
                     'Validator': lambda value: isinstance(value, RockGrading),
                     'Correct': 'RockGrading',
                     'Constant': True,
                     'Required': True,
                     'unit': '[-]'},

        'Dn50_core' : {'Default': None,
                       'Validator': lambda value: isinstance(value, int)
                                                  or isinstance(value, float),
                       'Correct': 'int or float',
                       'Constant': False,
                       'Required': True,
                       'unit': '[m]'},

        'safety' : {'Default': 1,
                    'Validator': lambda value: isinstance(value, int)
                                               or isinstance(value, float),
                    'Correct': 'int or float',
                    'Constant': True,
                    'Required': False,
                    'unit': '[-]'},

        'slope_toe' : {'Default': (2,3),
                       'Validator': lambda value: isinstance(value, tuple)
                                                  and len(value) == 2,
                       'Correct': 'tuple with length 2',
                       'Constant': False,
                       'Required': False,
                       'unit': '[-]'},

        'B_toe' : {'Default': None,
                   'Validator': lambda value: isinstance(value, int)
                                              or isinstance(value, float),
                   'Correct': 'int or float',
                   'Constant': False,
                   'Required': False,
                   'unit': '[m]'},

        'layers_underlayer' : {'Default': 2,
                               'Validator': lambda value: isinstance(value, int),
                               'Correct': 'int',
                               'Constant': True,
                               'Required': False,
                               'unit': '[-]'},

        'beta' : {'Default': 0,
                  'Validator': lambda value: isinstance(value, float)
                                             or isinstance(value, int),
                  'Correct': 'int or float',
                  'Constant': True,
                  'Required': False,
                  'unit': '[degrees]'},

        'Soil': {'Default': None,
                 'Validator': lambda value: isinstance(value, Soil),
                 'Correct': 'Soil',
                 'Constant': True,
                 'Required': False,
                 'unit': '[-]'},

        'phi': {'Default': 40,
                'Validator': lambda value: isinstance(value, float)
                                           or isinstance(value, int),
                'Correct': 'int or float',
                'Constant': True,
                'Required': False,
                'unit': '[degrees]'},

    }

    # only vkwargs if type of structure is rock rubble mound
    vkwargs_RRM = {
        'N' : {'Default': None,
               'Validator': lambda value: isinstance(value, int),
               'Correct': 'int',
               'Constant': True,
               'Required': True,
               'unit': '[-]'},

        'layers_rock' : {'Default': 2,
                         'Validator': lambda value: isinstance(value, int),
                         'Correct': 'int',
                         'Constant': True,
                         'Required': False,
                         'unit': '[-]'},

        'vdm' : {'Default': 'max',
                 'Validator': lambda value: isinstance(value, str)
                                            and value in ['min', 'max', 'avg'],
                 'Correct': 'str and {min, max, avg}',
                 'Constant': True,
                 'Required': False,
                 'unit': '[-]'},
    }

    # only vkwargs if type of structure is concrete rubble mound
    vkwargs_CRM = {
        'ArmourUnit' : {'Default': None,
                        'Validator': lambda value: isinstance(value, ConcreteArmour),
                        'Correct': 'ConcreteArmour or inherit from ConcreteArmour',
                        'Constant': True,
                        'Required': True,
                        'unit': '[-]'},

        'layers_units' : {'Default': 1,
                          'Validator': lambda value: isinstance(value, int),
                          'Correct': 'int',
                          'Constant': True,
                          'Required': False,
                          'unit': '[-]'},

        'filter_rule' : {'Default': None,
                         'Validator': lambda value: value in ['Rock', 'Xbloc', 'XblocPlus'],
                         'Correct': 'str and {Rock, Xbloc or XblocPlus}',
                         'Constant': True,
                         'Required': False,
                         'unit': '[-]'},

    }

    if type == 'Rock' or type == 'both':
        vkwargs.update(vkwargs_RRM)

    if type == 'ArmourUnit' or type == 'both':
        vkwargs.update(vkwargs_CRM)

    return vkwargs

def _C_vkwargs(type):
    """ Valid kwargs table for a vertical (composite) breakwater

    Construct a dictionary with all the valid kwargs for a vertical
    (composite) breakwater

    Parameters
    ----------
    type : {'Rock', 'ArmourUnit', 'both'}
        type of rubble mound breakwater, R for rock and C for concrete
        armour units.

    Returns
    -------
    vkwargs : dict
        dictionary with all the valid kwargs for a vertical (composite)
        breakwater
    """
    vkwargs = {
        'Pc': {'Default': None,
               'Validator': lambda value: isinstance(value, float)
                                          or isinstance(value, int),
               'Correct': 'int or float',
               'Constant': False,
               'Required': True,
               'unit': '[-]'},

        'rho_c': {'Default': None,
                   'Validator': lambda value: isinstance(value, float)
                                              or isinstance(value, int),
                   'Correct': 'int or float',
                   'Constant': True,
                   'Required': True,
                   'unit': '[kg/m³]'},

        'rho_fill': {'Default': None,
                     'Validator': lambda value: isinstance(value, float)
                                                or isinstance(value, int),
                     'Correct': 'int or float',
                     'Constant': True,
                     'Required': True,
                     'unit': '[kg/m³]'},

        'rho_w' : {'Default': None,
                   'Validator': lambda value: isinstance(value, float)
                                              or isinstance(value, int),
                   'Correct': 'int or float',
                   'Constant': True,
                   'Required': True,
                   'unit': '[kg/m³]'},

        'Bm': {'Default': None,
               'Validator': lambda value: isinstance(value, float)
                                       or isinstance(value, int),
               'Correct': 'int or float',
               'Constant': False,
               'Required': True,
               'unit': '[m]'},

        'hb': {'Default': None,
               'Validator': lambda value: isinstance(value, float)
                                       or isinstance(value, int),
               'Correct': 'int or float',
               'Constant': False,
               'Required': True,
               'unit': '[m]'},

        'LimitState' : {'Default': None,
                        'Validator': lambda value: isinstance(value, LimitState),
                        'Correct': 'LimitState',
                        'Constant': True,
                        'Required': True,
                        'unit': '[-]'},

        'slope_foreshore': {'Default': None,
                            'Validator': lambda value: isinstance(value, tuple)
                                                       and len(value) == 2,
                            'Correct': 'tuple with length 2',
                            'Constant': True,
                            'Required': True,
                            'unit': '[-]'},

        'mu': {'Default': None,
               'Validator': lambda value: isinstance(value, float),
               'Correct': 'float',
               'Constant': True,
               'Required': True,
               'unit': '[-]'},

        'safety' : {'Default': 1,
                    'Validator': lambda value: isinstance(value, int)
                                               or isinstance(value, float),
                    'Correct': 'int or float',
                    'Constant': True,
                    'Required': False,
                    'unit': '[-]'},

        'SF_sliding': {'Default': 1.2,
                       'Validator': lambda value: isinstance(value, int)
                                                  or isinstance(value, float),
                       'Correct': 'int or float',
                       'Constant': True,
                       'Required': False,
                       'unit': '[-]'},

        'SF_turning': {'Default': 1.2,
                       'Validator': lambda value: isinstance(value, int)
                                                  or isinstance(value, float),
                       'Correct': 'int or float',
                       'Constant': True,
                       'Required': False,
                       'unit': '[-]'},

        'beta': {'Default': 0,
                 'Validator': lambda value: isinstance(value, int)
                                            or isinstance(value, float),
                 'Correct': 'int or float',
                 'Constant': True,
                 'Required': False,
                 'unit': '[degrees]'},

        'slope_foundation' : {'Default': (2,3),
                              'Validator': lambda value: isinstance(value, tuple)
                                                         and len(value) == 2,
                              'Correct': 'tuple with length 2',
                              'Constant': False,
                              'Required': False,
                              'unit': '[-]'},

        'lambda_': {'Default': [1,1,1],
                    'Validator': lambda value: isinstance(value, list)
                                               and len(value) == 3,
                    'Correct': 'list with length 3',
                    'Constant': True,
                    'Required': False,
                    'unit': '[-]'},

        'filter_rule' : {'Default': None,
                         'Validator': lambda value: value in ['Rock', 'Xbloc', 'XblocPlus'],
                         'Correct': 'str and {Rock, Xbloc or XblocPlus}',
                         'Constant': True,
                         'Required': False,
                         'unit': '[-]'},

        'Grading' : {'Default': None,
                     'Validator': lambda value: isinstance(value, RockGrading),
                     'Correct': 'RockGrading',
                     'Constant': True,
                     'Required': True,
                     'unit': '[-]'},

        'pe_max' : {'Default': 500,
                    'Validator': lambda value: isinstance(value, int)
                                               or isinstance(value, float),
                    'Correct': 'int or float',
                    'Constant': True,
                    'Required': False,
                    'unit': '[kPa]'},

        'Soil': {'Default': None,
                 'Validator': lambda value: isinstance(value, Soil),
                 'Correct': 'Soil',
                 'Constant': True,
                 'Required': False,
                 'unit': '[-]'},
    }

    # only vkwargs if type of structure is rock foundation for the caisson
    vkwargs_RC = {

        'layers_rock' : {'Default': 2,
                         'Validator': lambda value: isinstance(value, int),
                         'Correct': 'int',
                         'Constant': True,
                         'Required': False,
                         'unit': '[-]'},
    }

    # only vkwargs if type of structure is armour unit foundation for caisson
    vkwargs_CC = {

        'layers_units' : {'Default': 1,
                    'Validator': lambda value: isinstance(value, int),
                    'Correct': 'int',
                    'Constant': True,
                    'Required': False,
                    'unit': '[-]'},

        'BermMaterial': {'Default': None,
                         'Validator': lambda value: isinstance(value, RockGrading)
                                                    or isinstance(value, ConcreteArmour),
                         'Correct': 'RockGrading or ConcreteArmour',
                         'Constant': True,
                         'Required': True,
                         'unit': '[-]'},

    }

    if type == 'Rock' or type == 'both':
        vkwargs.update(vkwargs_RC)

    if type == 'ArmourUnit' or type == 'both':
        vkwargs.update(vkwargs_CC)

    return vkwargs

def _validate_kwargs(params, vkwargs):
    """ Validate if all required input is given

    Parameters
    ---------
    params : dict
        dictionary with the processed kwargs
    vkwargs : dict
        dictionary with the valid kwargs, default values and a validator
    """
    for key, value in params.items():
        required = vkwargs[key]['Required']
        if required and value is None:
            raise InputError(
                f'{key} is a required parameter for the specified structure')
        else:
            continue

def _process_varying_kwarg(key, value, correct):
    """" Process varying kwarg

    Parameters
    ----------
    key : str
        name of the kwarg
    value : any
        value of the kwarg
    correct : str
        correct format of the parameter if not varying

    Returns
    -------
    value : any
        processed value
    """
    # set bool for converting slope
    convert_slope_to_tuple = False

    if isinstance(value, tuple) and len(value) == 3:
        # parameter is a tuple and has length of 3
        # check if value[0] and value[1] are tuples
        if isinstance(value[0], tuple) and isinstance(value[1], tuple):
            # parameter is a slope thus preprocessing is needed
            # tuples must be converted to floats (angles)
            lower = value[0][0]/value[0][1]
            upper = value[1][0]/value[1][1]

            # set new format of value for varying slopes
            value = (lower, upper, value[2])

            # set bool to convert values back to tuples
            convert_slope_to_tuple = True

        # check if format is correct (min, max, num)
        if value[1] > value[0] and isinstance(value[2], int):
            # make linspace of value and return the linspace
            varying_values = np.linspace(value[0], value[1], value[2])

            # check if varying value must be converted
            if convert_slope_to_tuple:
                converted_slopes = []

                # convert values to tuple
                for varying_val in varying_values:
                    converted_slopes.append((1, 1/varying_val))

                # convert the list of converted slopes to a ndarray
                dt = np.dtype('int, float')
                varying_values = np.array(converted_slopes, dtype=dt)

            # return the varying values
            return varying_values
        else:
            # incorrect format of min, max, num
            raise InputError(
                (f'{key} is allowed to vary but tuple must be formatted as '
                '(min, max, num)'))
    else:
        # incorrect type of length
        given_type = type(value).__name__
        given_len = len(value)
        raise InputError(
            (f'{key} is allowed to vary but must be tuple with length 3, is '
             f'{given_type} with length {given_len}. Don\'t want a varying '
             f'parameter? Specify argument as {correct}'))

def _process_kwargs(kwargs, vkwargs):
    """ Process the kwargs

    Check if the input kwarg is in the valid kwargs table, verify that
    a keyword is valid and that the value has the correct type.
    Furthermore, validate if all kwargs for the type of structure are
    given.

    Parameters
    ----------
    kwargs : dict
        dictionary with kwargs and their values
    vkwargs : dict
        dictionary with the valid kwargs, default values and a validator

    Returns
    -------
    dict
        dictionary with the kwargs and their values, and default values
    """
    # initialize params from vkwargs
    params = {}
    for key, value in vkwargs.items():
        params[key] = value['Default']

    # validate the kwargs, and replace the value in params
    # if the kwarg is a valid kwarg
    for key in kwargs.keys():
        if key not in vkwargs:
            # invalid kwarg of current structure, do not raise error
            # kwarg might be for another type of structure
            continue
        else:
            value = kwargs[key]
            valid = vkwargs[key]['Validator'](value)
            if valid:
                # argument has valid format, thus add to params
                params[key] = value
            else:
                # argument is invalid, thus raise error
                incorrect = type(value).__name__
                correct = vkwargs[key]['Correct']

                # check if parameter is allowed to vary
                if not vkwargs[key]['Required'] and value is None:
                    params[key] = value
                elif not vkwargs[key]['Constant']:
                    # parameter is allowed to vary
                    processed_value = _process_varying_kwarg(key, value, correct)

                    params[key] = processed_value

                else:
                    # parameter not allowed to vary thus incorrect input
                    raise InputError(
                        (f'{incorrect} is invalid for kwarg {key}, must be '
                         f'{correct}'))

    # validate if all required parameters are given
    _validate_kwargs(params, vkwargs)

    return params



import numpy as np

from .exceptions import InputError

def _RM_vparams():
    """ Parameters of RM that can exported to the design explorer """
    vparams = {
        'B': {'param': 'B',
              'key': 'df'},

        'Dn50_core': {'param': 'Dn50_core',
                      'key': 'df'},

        'B_toe': {'param': '_B_toe',
                  'key': 'attr',
                  'nested': False},

        'Rc': {'param': 'Rc',
               'key': 'attr',
               'nested': False},

        'slope': {'param': '_input_arguments',
                  'key': 'attr',
                  'nested': True},

        'slope_toe': {'param': '_input_arguments',
                      'key': 'attr',
                      'nested': True},

        'computed Dn50 armour': {'param': 'computed Dn50',
                                 'key': 'armour'},

        'class armour': {'param': 'class',
                         'key': 'armour'},

        'class Dn50 armour': {'param': 'class Dn50',
                              'key': 'armour'},

        'computed Dn50 underlayer': {'param': 'computed Dn50',
                                     'key': 'underlayer'},

        'class underlayer': {'param': 'rock class',
                             'key': 'underlayer'},

        'class Dn50 underlayer': {'param': 'class Dn50',
                                  'key': 'underlayer'},

        'computed Dn50 filter': {'param': 'computed Dn50',
                                     'key': 'filter layer'},

        'class filter': {'param': 'rock class',
                         'key': 'filter layer'},

        'class Dn50 filter': {'param': 'class Dn50',
                              'key': 'filter layer'},

    }

    return vparams

def _C_vparams():
    """ Parameters of C that can be exported to the design explorer """
    vparams = {
        'Pc': {'param': 'Pc',
               'key': 'caisson'},

        'hb': {'param': 'hb',
               'key': 'caisson'},

        'h_acc': {'param': 'h_acc',
                  'key': 'caisson'},

        'Rc': {'param': 'Rc',
               'key': 'caisson'},

        'B': {'param': 'B',
               'key': 'caisson'},

        'Bm': {'param': 'Bm',
               'key': 'caisson'},

        'slope_foundation': {'param': '_input_arguments',
                             'key': 'attr',
                             'nested': True},

        'computed Dn50 armour': {'param': 'computed Dn50',
                                 'key': 'armour'},

        'class armour': {'param': 'class',
                         'key': 'armour'},

        'class Dn50 armour': {'param': 'class Dn50',
                              'key': 'armour'},

        'computed Dn50 underlayer': {'param': 'computed Dn50',
                                     'key': 'underlayer'},

        'class underlayer': {'param': 'rock class',
                             'key': 'underlayer'},

        'class Dn50 underlayer': {'param': 'class Dn50',
                                  'key': 'underlayer'},

        'UC': {'param': 'geotechnical',
               'key': 'attr',
               'nested': False}
    }

    return vparams

def _DE_params(
        args, variant, row, concept, structure, slopes, change_CRM_class):
    """ Get the parameters for the design explorer

    Parameters
    ----------
    args : list
        list with arguments
    variant : dict
        a variant of a concept of a type of breakwater
    row : pd.Series
        current row of the df
    concept : obj
        a breakwater concept
    structure : {'RRM', 'CRM', 'RC', 'CC'}
        type of structure for export
    slopes : {angles, tuples}, optional, default: angles
        how the slopes must be exported. tuples will export the
        slope as (V,H) and angle as an angle in degrees
    change_CRM_class : bool
        True if the class of the armour layer of a CRM breakwater must
        be changed to a string, False if it must remain a float

    Returns
    -------
    dict
        dictionary with the parameter and value

    Raises
    ------
    InputError
        if the arg is not a valid argument for the design explorer
    """
    # set output dict
    to_explorer = {}

    # get the valid parameters
    if structure == 'RRM' or structure == 'CRM':
        vparams = _RM_vparams()
    elif structure == 'RC' or structure == 'CC':
        vparams = _C_vparams()
    else:
        raise NotImplementedError(f'{structure} is not supported')

    for arg in args:
        if arg in vparams:
            # get parameter and key for variant (from structure)
            param = vparams[arg]['param']
            key = vparams[arg]['key']

            # get the desired value for design explorer
            if key == 'df':
                # value is in the df
                value = row[param]

            elif key == 'attr':
                # value is attribute of the concept
                # check if nested key
                if vparams[arg]['nested']:
                    # nested
                    value = getattr(concept, param)[arg]
                    param = arg
                else:
                    # not nested
                    # check if param is UC
                    if param == 'geotechnical':
                        value = getattr(concept, param)['UC']
                    else:
                        value = getattr(concept, param)

            else:
                # check if the key is part of the variant
                if key in variant:
                    # key is in variant, thus get value
                    value = variant[key][param]
                else:
                    # not all variants have a filter layer
                    value = 'None'

            # check if format of armour layer CRM must be changed
            if change_CRM_class and param == 'class':
                if not isinstance(value, str):
                    SUB = str.maketrans('3', '³')
                    unit = 'm3'.translate(SUB)
                    value = f'{np.round(value,3)} {unit}'
                else:
                    pass

            # add to the dict and round value
            if param in ['slope', 'slope_toe', 'slope_foundation']:
                if slopes == 'angles':
                    angle = np.arctan(value[0]/value[1]) * 180/np.pi
                    to_explorer[arg] = round(angle, 1)
                elif slopes == 'tuples':
                    to_explorer[arg] = [(round(value[0], 2), round(value[1], 2))]
                else:
                    raise InputError(
                        (f'method {slopes} is invalid, please use angles or '
                          'tuples'))
            elif isinstance(value, str):
                to_explorer[arg] = [value]
            elif value is None:
                to_explorer[arg] = [value]
            else:
                to_explorer[arg] = [round(value, 3)]
        else:
            # invalid arg was given for this structure
            # not raising error because might be for another structure
            # set value to 0 for correct DesignExplorer export
            to_explorer[arg] = [0]

    return to_explorer



import numpy as np
import matplotlib.pyplot as plt
from tabulate import tabulate

from .utils.exceptions import InputError, user_warning, NotSupportedError, RockGradingError
from .core.goda import Goda
from .core.overtopping import vertical
from .core.toe import toe_berm_stability
from .core.substructure import underlayer, _supported_armour_layers, layer_coefficient
from .core.scour import scour_protection


class Caisson:
    """ Design a (composite) vertical breakwater

    Makes a conceptual design of a vertical or composite vertical
    breakwater, with a caisson on a rubble mound foundation. The
    following computations are performed:

    - The necessary size of the armour layer of the foundation is
      designed with the modified Tanimoto formula (Takahashi, 2002).
    - The required stone size for the core of the foundation
    - The water depth in front of the caisson is computed based on the
      dimensions of the foundation and water depth
    - The crest freeboard is computed with the formulae from EurOtop
      (2018), :py:func:`vertical` is used, which automatically
      classifies the breakwater so that the correct formula is used.
    - The required width of the caisson is computed with the extended
      Goda formula (Takahasi, 2002).
    - The required width of the scour protection with Sumer and Fredsoe
      (2000). Note that a scour protection is only added if the width
      of the foundation is not sufficient.
    - If a Soil is specified the bearing capacity of the soil will also
      be checked with Brinch Hansen (1970).

    Parameters
    ----------
    Pc : float
        contribution of concrete to the total mass of the caisson.
        value between 0 and 1
    rho_c : float
        density of concrete [kg/m³]
    rho_fill : float
        density of the fill material [kg/m³]
    rho_w : float
        density of water [kg/m³]
    Bm : float
        width of the berm [m]
    hb : float
        height of the foundation layer [m]
    layers : int
        number of layers in the armour layer of the toe berm
    BermMaterial : obj
        should be a :py:class:`RockGrading` or armour unit class which
        inherits from :py:class:`ConcreteArmour`, for instance
        :py:class:`Xbloc` or :py:class:`XblocPlus`
    LimitState : :py:class:`LimitState` or list of :py:class:`LimitState`
        ULS, SLS or another limit state defined with
        :py:class:`LimitState`
    slope_foreshore : tuple
        slope of the foreshore (V, H). For example a slope of 1:100 is
        defined as (1,100)
    mu : float
        friction factor between the caisson and the foundation [-]
    safety : float, optional, default: 1
        safety factor of design (number of standard deviations from the
        mean)
    SF_sliding : float, optional, default: 1.2
        safety factor against sliding. Default value according to
        Goda (2000)
    SF_turning : float, optional, default: 1.2
        safety factor against sliding. Default value according to
        Goda (2000)
    beta : float, optional, default: 0
        angle between direction of wave approach and a line normal to
        the breakwater [degrees]
    slope_foundation : tuple, optional, default: (2,3)
        Slope of the armour layer (V, H). For example a slope of 2V:3H
        is defined as (2,3)
    lambda_ : list, optional, default: [1, 1, 1]
        modification factors of Takahasi (2002) for alternative
        monolithic breakwater. Input must be
        \\lambda_= [:math:`\\lambda_1, \\lambda_2, \\lambda_3`].
    pe_max : float, optional, default: 500
        maximum value of the bearing pressure at the heel of the caisson.
        Default value is set to 500 kPa, Goda (2000) advises a value
        between 400 and 500 kPa. [kPa]
    filter_rule : {'Rock', 'Xbloc', 'XblocPlus'}, optional, default: None
        filter rule to use for the substructure of the breakwater, for
        rock, Xbloc and XblocPlus the correct filter rule is
        automatically selected. In case another type of armour layer is
        used one of these filter rules must be chosen.
    Grading : :py:class:`RockGrading`, optional, default: None
        standard rock grading defined in the NEN-EN 13383-1 or a user
        defined rock grading. Required if the BermMaterial is not a
        :py:class:`RockGrading`.
    Soil : :py:class:`Soil`, optional, default: None
        by default Soil is None, which means that the geotechnical checks
        are not performed. By specifying a Soil object, the geotechnical
        checks are automatically performed.
    id : int, optional, default: None
        unique id of the breakwater

    Attributes
    ----------
    logger : dict
        dict of warnings and messages
    structure : dict
        dictionary with the computed Dn50, rock class, and average Dn50
        of the rock class for each layer and the toe. This dictionary
        includes all variants, use :py:meth:`get_variant` to get the
        parameters of one specific variant. Alternatively,
        :py:meth:`print_variant` can be used to print the details of
        one, multiple or all variants.
    id : int
        unique id of the breakwater
    variantIDs : list
        list with the IDs of the variants generated for this rubble
        mound breakwater.
    price : dict
        cost of each variant, generated when :py:meth:`cost` or
        :py:meth:`dry_dock` is used.
    width_scour : float
        the required length of the scour protection [m]. The distance
        is measured from the front of the caisson
    geotechnical : dict
        dictionary with the computed values from the Brinch Hansen
        equation, has keys: Fr with the resistance, N with the net
        downward force and UC, which is the unity check (N/Fr). This
        dictionary is generated when the :py:obj:`Soil` argument is
        specified.
    """

    def __init__(
            self, Pc, rho_c, rho_fill, rho_w, Bm, hb, layers, BermMaterial,
            LimitState, slope_foreshore, mu, safety=1, SF_sliding=1.2,
            SF_turning=1.2, beta=0, slope_foundation=(2,3), lambda_=[1,1,1],
            pe_max=500, filter_rule=None, Grading=None, Soil=None, id=None,
            **kwargs):
        """ See help(Caisson) for more info """
        # set logger and structure
        self.logger = {'INFO': [], 'WARNING': []}
        self.structure = {}

        # set id and variantIDs
        self.id = id
        self.variantIDs = ['a']

        # set cost Attribute
        self.price = None

        # compute relative buoyant density
        delta = (BermMaterial.rho - rho_w)/rho_w

        # convert beta from degrees to rad
        beta = beta*np.pi/180

        # compute the angle of the foreshore
        slope_foreshore = np.arctan(slope_foreshore[0]/slope_foreshore[1])

        # set top layer and filter layer
        top_layer = BermMaterial.name
        supported = _supported_armour_layers()

        if top_layer in supported:
            # supported armour layer
            filter_rule = top_layer
            if Grading is None and top_layer != 'Rock':
                # no grading for armour unit, thus raise error
                supported_armour_units = [
                    layer for layer in supported if layer is not 'Rock']
                support_out = ', '.join(supported_armour_units)
                raise InputError(
                    'Missing argument: Grading. The top layer is made out of '
                    f'{support_out}, therefore a RockGrading must be specified')
            elif top_layer == 'Rock':
                Grading = BermMaterial

        if filter_rule is None:
            supported_rules = ', '.join(supported)
            raise NotSupportedError(
                (f'Filter rule for {top_layer} is not implemented, set '
                 f'filter rule to use with filter_rule to {supported_rules}'))

        # restructure the input LimitState(s)
        self._LimitStates = []
        if isinstance(LimitState, list):
            self._LimitStates.extend(LimitState)
        else:
            self._LimitStates.append(LimitState)

        # set private attribute for plotting cross section
        self._input_arguments = {
            'Grading': Grading,
            'slope_foundation': slope_foundation,
            'armour': top_layer}

        # design the armour layer of the foundation
        # set temporary values to check changes
        Dn50_berm, state = 0, 0

        # iterate over the LimitStates
        for i, LimitState in enumerate(self._LimitStates):
            # get the wave height and submerged depth of caisson
            H13 = LimitState.get_Hs(definition='H13')
            T13 = LimitState['T13']
            h = LimitState.h
            d = h - hb

            # set values for the while loop
            Dn50 = 0
            compute_berm = True

            # compute Dn50 of the berm in a while loop,
            # because d = h - hb - layers*Dn50
            while compute_berm:
                Dn50_temp = toe_berm_stability(
                    Hs=H13, T=T13, d=d, Bm=Bm, Delta=delta, beta=beta)
                if (Dn50_temp - Dn50) > -0.05 and (Dn50_temp - Dn50) < 0.05:
                    compute_berm = False
                else:
                    Dn50 = Dn50_temp
                d = h - hb - layers*Dn50

            # check if computed Dn50 of current LimitState is larger
            # than current normative Dn50
            if Dn50 > Dn50_berm:
                # value is larger than current normative
                Dn50_berm = Dn50
                state = i

        if d <= 0:
            raise InputError(
                ('Encountered negative value for d. The chosen foundation '
                 'height (hb) is probably very close to the water depth (h)'))

        class_berm = BermMaterial.get_class(Dn50_berm)

        if top_layer == 'Rock':
            # berm material is made out of rock
            class_dn = BermMaterial.get_class_dn50(class_berm)
        else:
            # berm material is made out of concrete armour units
            class_dn = class_berm**(1/3)

        # compute d, the depth above the foundation
        d = self._LimitStates[state].h - hb - layers*Dn50_berm

        # add dimensions of armour layer to the structure
        self.structure['armour'] = {
            'computed Dn50': Dn50_berm,
            'class': class_berm,
            'class Dn50': class_dn,
            'state': state,
            'layers': layers}

        # check if additional layer is needed
        range_Dn50_u = underlayer(
            Dn_armour=class_dn, armour_layer=filter_rule,
            rho=BermMaterial.rho, rho_rock=Grading.rho)

        class_u, class_dn_u = [], []

        for dn_u in range_Dn50_u:
            rock_class = Grading.get_class(dn_u)

            if rock_class in class_u:
                # rock class is already in the underlayer so no need to
                # generate a new variant because design will be the same
                continue
            else:
                # first loop or new variant
                class_u.append(rock_class)
                class_dn_u.append(Grading.get_class_dn50(rock_class))

        self.structure['foundation'] = {'computed Dn50': range_Dn50_u,
                                        'class': class_u,
                                        'class Dn50': class_dn_u,
                                        'state': 'see armour'}

        # add message to the log if new variant was generated
        if len(class_u) == 2:
            self.logger['INFO'].append(
                'two rock classes possible for the underlayer, generated new '
                'variant b')
            self.variantIDs.append('b')

        # compute the crest height
        # set temporary values
        Rc, state_overtopping, B, m = 0, 0, 0, 0

        for i, LimitState in enumerate(self._LimitStates):
            self.logger['INFO'].append(f'computing with {LimitState.label}:')
            # get hydraulic parameters
            Hm0 = LimitState.get_Hs(definition='Hm0')
            L = LimitState.L(period='T_m_min_1', deep_water=True)
            s = LimitState.s(number='spectral')
            h = LimitState.h

            Rc_temp = vertical(
                Hm0=Hm0, q=LimitState['q'], h=h, d=d, L_m_min_1=L,
                s_m_min_1=s, safety=safety, logger=self.logger)

            # check if computed Rc of current LimitState is larger
            # than current normative Rc
            if Rc_temp > Rc:
                # if larger the normative Rc must be changed
                Rc = Rc_temp
                state_overtopping = i

            # compute/get wave heights for Goda
            H13 = LimitState.get_Hs(definition='H13')
            Hmax = LimitState['Hmax']

            # compute the height of the caisson
            h_acc = h - hb

            goda = Goda(
                Hs=H13, Hmax=Hmax, h=h, d=d, h_acc=h_acc, hc=Rc_temp, Bm=Bm,
                T=LimitState['T13'], beta=beta, rho=rho_w, logger=self.logger,
                slope_foreshore=slope_foreshore, lambda_=lambda_)

            B_temp = goda.required_width(
                Pc=Pc, rho_c=rho_c, rho_f=rho_fill, rho_w=rho_w, mu=mu,
                t=0.5, SF_sliding=SF_sliding, SF_turning=SF_turning,
                logger=self.logger)

            # check bearing capacity
            pe = goda.bearing_pressure(Pc=Pc, rho_c=rho_c, rho_fill=rho_fill)

            if pe/1000 > pe_max:
                # bearing capacity is not large enough
                B_temp = goda.bearing_pressure_width(
                    B1=B_temp, Pc=Pc, rho_c=rho_c, rho_fill=rho_fill,
                    pe_max=pe_max)

                # replace normative msg from required_width in logger
                self.logger['INFO'][-1] = ('The bearing pressure is normative'
                    ' for the computation of the width')

            m_temp = goda.mass(Pc=Pc, rho_c=rho_c, rho_fill=rho_fill)

            # check if values have changed from the current normative
            if m_temp > m:
                # set new normative values
                B = B_temp
                m = m_temp
                self.goda = goda
                state_goda = i

        # compute submerged depth of the caisson
        h_acc = self._LimitStates[state_goda].h - hb

        self.structure['caisson'] = {
            'hb': hb, 'h_acc': h_acc, 'Pc': Pc, 'd': d,
            'Rc': Rc, 'state_overtop': state_overtopping,
            'B': B, 'state_goda': state_goda, 'Bm': Bm}

        # Compute required scour protection
        self.width_scour = 0
        for LimitState in self._LimitStates:
            w = scour_protection(L=LimitState.L(period='Tm'))

            # check if larger than previous value
            if w >= self.width_scour:
                # set w as new width scour
                self.width_scour = w

        # make geotechnical design
        if Soil is not None:
            # compute the effective width of the foundation
            B_eff_sub = self._effective_width_foundation(rho_w, hb, B, m)

            # compute the forces of the foundation on the subsoil
            # compute horizontal stress per meter
            t = (self.goda.P()/(h_acc + Rc))/B_eff_sub/1000

            # compute vertical stress per meter
            p = (self.goda._dFv(m) + self._Fsill(rho_w, hb, B))/B_eff_sub/1000

            # compute the bearing capacity of the subsoil
            p_cap = Soil.brinch_hansen(
                p=p, t=t, B=B_eff_sub, L=None, q=0, rho_w=rho_w)

            # check if capacity is enough
            if p >= p_cap:
                # force is larger
                user_warning(
                    ('Bearing capacity of the soil is smaller than the '
                     'exerted stress'))

            # make geotechnical attribute and add values
            self.geotechnical = {'p_cap': p_cap, 'p': p, 'UC': p/p_cap}

    def _Fsill(self, rho_w, hb, B):
        """ Compute the downward force of the foundation """
        # slope
        V, H = 1, 1

        # get the grading
        Grading = self._input_arguments['Grading']

        # compute downward force of the foundation
        # only the foundation directly below the caisson
        return (Grading.rho - rho_w)*9.81*(V/H * hb**2 + B*hb)

    def _effective_width_foundation(self, rho_w, hb, B, m):
        """ Compute the effective width of the foundation """
        # slope
        V, H = 1, 1

        # compute moment around the middle of the foundation
        Mb = self.goda.Ma() + self.goda.P() * hb

        # compute the eccentricity
        e = Mb/(self.goda._dFv(m) + self._Fsill(rho_w, hb, B))

        # compute effective width of the foundation
        return B + 2*V/H*hb - 2*e

    def _validate_variant(self, variants):
        """ Validate the input of the variant

        Parameters
        ----------
        variants : tuple
            variantIDs given as args input
        """
        # check if a variant is specified
        if not variants:
            # no input is given so get the valid args for variant
            valid_args = self.variantIDs
            valid_args.append('all')

            # raise error
            valid = ', '.join(valid_args)
            raise InputError(
                'did not specify which variants to use, possible arguments '
               f'are {valid}')

        # check if input all is in variants
        if 'all' in variants:
            # set specified variants to all variantIDs
            variants = tuple(self.variantIDs)

        # return the variants
        return variants

    def _layers(self, variantID):
        """ compute the coordinates of all layers

        Parameters
        ----------
        variantID : str
            identifier of the variant, see :py:attr:`variantIDs` for a
            list of all generated variants.

        Returns
        -------
        dict
            coordinates of all layers
        """
        # get the structure of the current variant
        structure = self.get_variant(variantID)

        # set empty dict to store coordinates in
        coordinates = {}

        # get the slope
        V, H = self._input_arguments['slope_foundation']

        # conmpute constant to transfrom thickness of layer to x and
        # y coordinates, switched V and H because orthogonality
        transform_x = V/np.sqrt(V**2+H**2)
        transform_y = H/np.sqrt(V**2+H**2)

        # get geometrical parameters of the caisson
        B = structure['caisson']['B']
        hc = structure['caisson']['h_acc'] + structure['caisson']['Rc']
        hb = structure['caisson']['hb']
        Bm = structure['caisson']['Bm']

        # determine thickness of the armour layer
        layers_armour = structure['armour']['layers']
        kt_armour = layer_coefficient(
            self._input_arguments['armour'], layers=layers_armour,
            placement='standard')
        dn = structure['armour']['class Dn50']
        t_armour = kt_armour*layers_armour*dn

        # check if scour protection is required
        if Bm >= self.width_scour:
            # no scour protection required
            t_scour = 0
            req_width = 0

        else:
            # define thickness of the scour
            t_scour = 0.4

            # compute required width of the scour protection
            req_width = self.width_scour - Bm - H/V * (hb + t_armour - t_scour)

            if req_width <= 0:
                # negative or zero width required width of the foundation
                # is enough, thus set values to 0
                t_scour, req_width = 0, 0

        # define line of the caisson
        # clockwise starting at the lower left of the caisson
        coordinates['caisson'] = {
            'x': [-0.5*B, -0.5*B, 0.5*B, 0.5*B, -0.5*B],
            'y': [hb, hb + hc, hb + hc, hb, hb]
        }

        # define points of the foundation
        arm_top = hb + t_armour

        arm_x3 = -0.5*B
        arm_x2 = arm_x3 - Bm
        arm_x1 = arm_x2 - H/V * (arm_top - t_scour)

        found_x0 = arm_x1 - req_width - H*t_scour/V
        found_x1 = arm_x1 - req_width
        found_x2 = (arm_x1 + H*(t_armour * transform_y)/V
                       + t_armour*transform_x)
        found_x3 = found_x2 + H/V * (hb - t_scour)
        found_x4 = -0.5*B
        found_x5 = 0.5*B + Bm
        found_x6 = found_x5 + H/V * hb

        # define line of the foundation layers
        # clockwise starting at the left with the upper line
        coordinates['armour'] = {
            'x': [arm_x1, arm_x2, arm_x3, arm_x3, found_x3, found_x2, arm_x1],
            'y': [t_scour, arm_top, arm_top, hb, hb, t_scour, t_scour]
        }

        # check if scour protection is needed
        if req_width != 0:
            # protection needed
            coordinates['foundation'] = {
                'x': [found_x0, found_x1, found_x2, found_x3, found_x4,
                      found_x5, found_x6, found_x0],
                'y': [0, t_scour, t_scour, hb, hb,
                      hb, 0, 0]
            }
        else:
            # not needed
            coordinates['foundation'] = {
                'x': [found_x2, found_x3, found_x4, found_x5, found_x6,
                      found_x2],
                'y': [0, hb, hb, hb, 0,
                      0]
            }

        return coordinates

    def print_logger(self, level='warnings'):
        """ Print messages and warnings in the logger

        Parameters
        ----------
        msg_level : {'info', 'warnings'}, optional, default: 'warnings'
            specify print level, highest level is warnings and lowest
            level is info. Note that the info level will also print all
            warnings
        """
        # check if correct input has been given
        if level.lower() not in ['warnings', 'info']:
            raise NotSupportedError(
                f'{level} not implemented, must be info or warnings')

        # print logger
        for type, messages in self.logger.items():
            if level.lower() == 'warnings':
                if type == 'INFO':
                    continue
            print(f'{type}:')
            if messages:
                for message in messages:
                    print(message)
            else:
                print(f'no {type.lower()} messages in log')
            print()

    def get_variant(self, variantID):
        """ Get the dimensions for the specified variant

        Parameters
        ----------
        variantID : str
            identifier of the variant, see :py:attr:`variantIDs` for a
            list of all generated variants.

        Returns
        -------
        dict
            Parameters and values of the caisson (B, Rc) and the
            foundation layer (Dn50, Rock class) for one variant.

        Raises
        ------
        KeyError
            If there is no variant with the given identifier
        """
        variant = {}
        if variantID in self.variantIDs:
            key_u = self.variantIDs.index(variantID)
        else:
            raise KeyError(f'Variant with ID = {variantID} is not a variant, '
                           f'generated variants are: {self.variantIDs}')

        # add caisson
        variant['caisson'] = self.structure['caisson']

        # add foundation armour and underlayer
        variant['armour'] = self.structure['armour']

        if 'foundation' in self.structure:
            variant['foundation'] = {}
            for param, val in self.structure['foundation'].items():
                if param == 'state':
                    variant['foundation'][param] = val
                else:
                    variant['foundation'][param] = val[key_u]

        # return the generated variant
        return variant

    def print_variant(self, *variants, decimals=3):
        """ Print the details for the specified variant(s)

        This method will print the details of the structure for
        the specified variant(s). It prints the dimensions of the
        caisson (B, Rc) and the dimensions of the foundation layer
        (Dn50, rock class). Furthermore, from the table the normative
        LimitState for the design can be read.

        Parameters
        ----------
        *variants : str
            IDs of the variants to plot, see :py:attr:`variantIDs` for
            a list of all generated variants. If 'all' is in the
            arguments, all variants will be plotted.
        decimals : int, optional, default: 3
            number of decimal places to round to

        Raises
        ------
        InputError
            If no arguments are specified
        KeyError
            If there is no variant with the given identifier
        """
        # validate variants
        variants = self._validate_variant(variants)

        # print variant
        for id in variants:
            variant = self.get_variant(id)

            # print the name of the table
            if isinstance(self.id, int):
                table_name = f'Variant {self.id}{id}:'
            else:
                table_name = f'Variant {id}:'

            # set defaults and generate empty tables
            headers_caisson = ['parameter', 'value']
            table_caisson, table_f = [], []
            set_headers = True

            # fill the empty tables
            for i, (layer, dimensions) in enumerate(variant.items()):
                # caisson table
                if layer == 'caisson':
                    for param, val in dimensions.items():
                        if param == 'Rc':
                            val = np.round(val, decimals)
                            state = self.structure['caisson']['state_overtop']
                            label = self._LimitStates[state].label
                            table_val = f'{val} (with {label})'
                            table_caisson.append([param, table_val])
                        elif param == 'B':
                            val = np.round(val, decimals)
                            state = self.structure['caisson']['state_goda']
                            label = self._LimitStates[state].label
                            table_val = f'{val} (with {label})'
                            table_caisson.append([param, table_val])
                        elif param == 'state_overtop' or param == 'state_goda':
                            continue
                        else:
                            table_caisson.append(
                                [param, np.round(val, decimals)])
                # table for the foundation layer
                else:
                    if set_headers:
                        # set headers in first iteration
                        headers_foundation = list(dimensions.keys())
                        headers_foundation.insert(0, 'layer')
                        set_headers = False

                    # make row with the layer and dimensions
                    row = [layer]
                    row.extend(dimensions.values())

                    # check if state is in the headers
                    if 'state' in headers_foundation:
                        # get the normative state and index of state
                        state = dimensions['state']
                        i_state = headers_foundation.index('state')

                        # check if state is an int
                        if isinstance(state, int):
                            # get the label of the normative limitstate
                            label = self._LimitStates[state].label
                            row[i_state] = label

                    # add row to the table
                    table_f.append(row)

            # print tables
            print(table_name)
            print()
            print('  Caisson dimensions')
            print(tabulate(table_caisson, headers_caisson, tablefmt="github"))
            print()
            print('  Foundation dimensions')
            print(tabulate(
                table_f, headers_foundation, tablefmt="github",
                floatfmt=(f'.{decimals}f')))
            print('\n')

    def plot(self, *variants, wlev=None, save_name=None):
        """ Plot the cross section of the specified breakwater(s)

        Parameters
        ----------
        *variants : str
            IDs of the variants to plot, see :py:attr:`variantIDs` for
            a list of all generated variants. If 'all' is in the
            arguments, all variants will be plotted.
        wlev : str, optional, default: None
            label of the :py:class:`LimitState` from which the water
            level will be plotted. If no value is specified the water
            level from the normative limit state is used, which is the
            normative LimitState from the crest freeboard computation.
        save_name : str, optional, default: None
            if given the cross section is not shown but saved with the
            given name

        Raises
        ------
        InputError
            If no variants are specified or if the label of wlev is not
            a valid label of a :py:class:`LimitState`
        KeyError
            If there is no variant with the given identifier
        """
        # validate variants
        variants = self._validate_variant(variants)

        if wlev is None:
            wlev = self.structure['caisson']['state_overtop']
        else:
            for i, LimitState in enumerate(self._LimitStates):
                if LimitState.label == wlev:
                    wlev = i
                    break

        # check if wlev has been changed from LimitState label into index
        if isinstance(wlev, str):
            # wlev is still a string so not changed, which means that
            # the specified wlev is not a specified LimitState
            raise InputError('There is no LimitState with the given label')

        # set figure
        plt.figure(figsize=(10, 5))

        for i, id in enumerate(variants):
            if len(variants) == 2:
                # make subplot if two variants must be plotted
                plt.subplot(1, 2, i+1)

            # get the coordinates
            coordinates = self._layers(id)

            # set xlim_max and xlim_min variable
            xlim_max, xlim_min = 0, 0

            # plot lines
            for layer, lines in coordinates.items():
                plt.plot(lines['x'], lines['y'], color='k')

                # check largest value for xlim
                if np.max(lines['x']) >= xlim_max:
                    # set max as xlim_max
                    xlim_max = np.max(lines['x'])

                # check smallest value for xlim
                if np.min(lines['x']) <= xlim_min:
                    # set min as xlim_min
                    xlim_min = np.min(lines['x'])

            # plot bottom and wlev
            x_wlev_max = np.max(coordinates['caisson']['x'])

            plt.axhline(y=0, color='k', linewidth=2)
            plt.hlines(
                y=self._LimitStates[wlev].h, xmin=xlim_min*1.2,
                xmax=-x_wlev_max, color='b')
            plt.hlines(
                y=self._LimitStates[wlev].h, xmin=x_wlev_max,
                xmax=xlim_max*1.2, color='b')

            # set xlim and ylim
            ymax = np.max(coordinates['caisson']['y'])*1.2
            plt.xlim(xlim_min*1.2, xlim_max*1.2)
            plt.ylim(-0.5, ymax)

            # add title to the plot
            if save_name is None:
                if isinstance(self.id, int):
                    title = ('Cross section of monolithic breakwater '
                             f'{self.id}{id}')
                else:
                    title = f'Cross section of monolithic breakwater {id}'
            else:
                name = save_name.split('/')[-1]
                title = f'Cross section of {name}'

            # add title, grid and set equal aspect ratio
            plt.title(title)
            plt.grid()
            plt.gca().set_aspect('equal', adjustable='box')

        plt.tight_layout()

        # save the figure
        if save_name is not None:
            plt.savefig(f'{save_name}.png')
            plt.close()
        else:
            plt.show()

    def plot_pressure(self):
        """ Plot pressure distribution computed extended Goda formula

        Plots the pressure distribution computed with the extended
        Goda formula (Takahasi, 2002) together with the dimensions
        of the monolithic breakwater.

        .. warning::
           Do not read the dimensions of the monolithic breakwater from
           the axes of the figure. The correct dimensions of the
           monolithic breakwater are depicted in the figure, or use
           :py:meth:`get_variant` or :py:meth:`print_variant`.
        """
        self.goda.plot()

    def area(self, variantID):
        """ Compute the area of all layers

        Method computes the area of each layer using Gauss's area
        formula. Which is given by the following formula:

        .. math::
           \\mathbf{A}=\\frac{1}{2} | \\sum_{i=1}^{n-1} x_{i} y_{i+1}
           +x_{n} y_{1}-\\sum_{i=1}^{n-1} x_{i+1} y_{i}-x_{1} y_{n} |

        Parameters
        ----------
        variantID : str
            identifier of the variant, see :py:attr:`variantIDs` for a
            list of all generated variants.

        Returns
        -------
        dict
            dict with the area of each layer
        """
        # get the coordinates of the layers
        coordinates = self._layers(variantID)

        # iterate over the layers to compute the area
        area = {}
        for layer, coord in coordinates.items():
            # get the x and y coordinates
            x = coord['x']
            y = coord['y']

            # use Gauss's area formula
            A = 0.5*np.abs(np.dot(x,np.roll(y,1))-np.dot(y,np.roll(x,1)))

            # add to area dict
            area[layer] = A

        return area

    def cost(
            self, *variants, concrete_price, fill_price, unit_price=None,
            output='variant'):
        """ Compute the cost per meter for each variant

        Method to compute the cost of each generated variant, the cost
        is computed per meter

        Parameters
        ----------
        *variants : str
            IDs of the variants to plot, see :py:attr:`variantIDs` for
            a list of all generated variants. If 'all' is in the
            arguments, all variants will be plotted.
        concrete_price : float
            price of concrete per m³
        fill_price : float
            price of the fill material per m³
        unit_price : float, optional, default: None
            the cost of an armour unit per m³, required if the armour
            layer of the foundation is made out of armour units
        output : {variant, layer, average}
            format of the output dict, variant returns the total cost
            of each variant, layer the cost of each layer for each
            variant and average returns the average cost.

        Returns
        -------
        dict
            the cost

        Raises
        ------
        RockGradingError
            if no pricing is included in the given RockGrading
        """
        # validate variants
        variants = self._validate_variant(variants)

        # get the grading, and check if the cost has been added
        Grading = self._input_arguments['Grading']

        if 'price' in Grading[list(Grading.grading.keys())[0]]:
            # pricing has been added
            pass
        else:
            # pricing has not been added, raise error
            raise RockGradingError('There is no pricing in the RockGrading')

        # set empty dict to store the output in
        cost = {}

        # iterate over the generated variants
        for id in variants:
            # get the areas and structure of the variants
            areas = self.area(id)
            structure = self.get_variant(id)

            # iterate over the layers to price each layer
            variant_price = {}
            for layer, area in areas.items():
                if layer is 'caisson':
                    # compute price of the caisson
                    Pc = structure['caisson']['Pc']
                    price = area*Pc*concrete_price + area*(1-Pc)*fill_price
                elif (self._input_arguments['armour'] is not 'Rock'
                        and layer is 'armour'):
                    # concrete armour units
                    if unit_price is not None:
                        price = area * unit_price
                    else:
                        raise InputError(
                            ('argument unit_price is required when computing '
                             'the cost with armour units as BermMaterial'))
                else:
                    # layer of the breakwater
                    rock_class = structure[layer]['class']

                    # get the price per meter
                    price = Grading[rock_class]['price'] * area

                # add to dict
                variant_price[layer] = np.round(price, 2)

            # add to cost dict
            if output is 'variant' or output is 'average':
                # add total cost of all layers
                cost[id] = np.round(np.sum(list(variant_price.values())), 2)
            elif output is 'layer':
                # add the cost of each layer
                cost[id] = variant_price
            else:
                # invalid input
                raise NotSupportedError(
                    (f'Cost can\'t be exported as {output}, must be variant, '
                      'layer or average'))

        # check if average must be computed
        if output is 'average':
            # compute average cost
            cost = {'average': np.round(np.average(list(cost.values())), 2)}

        # check if dry dock has been added
        if self.price is None:
            # no dry dock added
            self.price = cost
        else:
            # check output
            if output == 'variant':
                # add investment to each variant
                for id, price in cost.items():
                    cost[id] = price + self.price

            elif output == 'layer':
                # add investment as a key
                cost['investment'] = self.price

            elif output == 'average':
                # add investment
                cost['average'] += self.price

        # check if the average cost is computed for only one variant
        if len(variants) == 1 and output == 'average':
            # print user_warning and change key into variant
            cost[variants[0]] = cost.pop('average')
            user_warning(
                ('Computing the average for one variantID, changed key '
                 'average in dict with the specified variantID'))

        return cost

    def dry_dock(self, investment, length):
        """ Add the investment cost of a dry dock to the concept

        This method adds the investment required to rent a dry dock to
        the concept. The investment cost is added to the concept by
        dividing the investment trough the length of the breakwater to
        get the required investment per running meter.

        Parameters
        ----------
        investment : float
            the investment required to rent a dry dock
        length : float
            length of the breakwater [m]
        """
        # check if cost have already been added
        if self.price is not None:
            # iterate over the price dict
            for id, price in self.price.items():
                # check if mode was average
                if id == 'average':
                    # mode was average
                    self.price[id] = price + investment/length
                    break
                else:
                    # check if price is a dict
                    if isinstance(price, dict):
                        # mode was layer
                        # add investment as additional key
                        self.price['investment'] = investment/length
                        break
                    else:
                        # set new price
                        self.price[id] = price + investment/length

        else:
            # set investment as attribute
            self.price = investment/length
